#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import copy
import gzip
import os
import shlex
import shutil
import subprocess
import sys
import tarfile
import tempfile
import textwrap
import time
import urllib
import urllib.error
import urllib.parse
import urllib.request
import uuid
import zipfile
from collections.abc import Sequence
from concurrent.futures import ProcessPoolExecutor
from functools import wraps
from typing import Any, Callable, Dict, List, Tuple, Union

from tqdm import tqdm as ProgressBar

from .controller import send_message_to_controller
from .eval import interpolate
from .messages import decode_msg, encode_msg
from .parser import SoS_Script
from .syntax import SOS_ACTION_OPTIONS
from .targets import executable, file_target, path, paths, sos_targets
from .utils import (StopInputGroup, TerminateExecution, TimeoutInterProcessLock,
                    env, fileMD5, get_traceback, load_config_files, short_repr,
                    textMD5, transcribe)

__all__ = [
    "SoS_Action",
    "script",
    "sos_run",
    "fail_if",
    "warn_if",
    "stop_if",
    "download",
    "run",
    "perl",
    "report",
    "pandoc",
]


def get_actions() -> List[Any]:
    # get the name of all actions, which are identified by an attribute
    # run_mode of the function
    return [k for k, v in globals().items() if hasattr(v, "run_mode")]


#
# A decoration function that allows SoS to replace all SoS actions
# with a null action. Option run_mode is deprecated and might be
# removed later on.
#


def SoS_Action(
    run_mode: Union[str, List[str]] = "deprecated",
    acceptable_args: Union[Tuple[str], List[str]] = ("*",),
    default_args: Dict[str, Dict[str, str]] = {},
) -> Callable:

    def runtime_decorator(func):

        @wraps(func)
        def action_wrapper(*args, **kwargs):
            # if container in args, a large number of docker-specific
            # args would be allowed.
            for k in default_args:
                if k in default_args and k not in kwargs:
                    kwargs[k] = default_args[k]
            if "*" not in acceptable_args and all(
                    x not in kwargs for x in ("docker_image", "container",
                                              "template", "template_name")):
                for key in kwargs.keys():
                    if key not in acceptable_args and key not in SOS_ACTION_OPTIONS:
                        raise ValueError(
                            f'Unrecognized option "{key}" for action {func}')
            # docker files will be downloaded in run or prepare mode
            # this option is independent of container...
            if "docker_file" in kwargs and env.config["run_mode"] in [
                    "run",
                    "interactive",
            ]:
                from .docker.client import SoS_DockerClient

                docker = SoS_DockerClient()
                docker.load_image(kwargs["docker_file"])
            # handle image
            if "docker_image" in kwargs:
                if "container" in kwargs and kwargs["container"]:
                    raise ValueError(
                        "Option docker_image is deprecated and should not be specified with option container"
                    )
                kwargs["container"] = "docker://" + kwargs["container"]
            if "container" in kwargs and kwargs["container"]:
                if not isinstance(kwargs["container"], str):
                    raise ValueError(
                        f'A string in the format of "scheme://tag" is expected for option container, {kwargs["container"]} provided'
                    )
                engine = (
                    kwargs["engine"]
                    if "engine" in kwargs and kwargs["engine"] else None)
                if "://" in kwargs["container"]:
                    cty, cname = kwargs["container"].split("://", 1)
                elif kwargs["container"].endswith(
                        ".simg") or kwargs["container"].endswith(".sif"):
                    engine = "singularity"
                    cty = "file"
                    cname = kwargs["container"]
                else:
                    cty = None
                    cname = kwargs["container"]
                # now let us figure out image and engine
                # if engine is specified
                if engine == "docker":
                    if cty is not None and cty != "docker":
                        raise ValueError(
                            f"docker engine only allows docker container {cty} specified"
                        )
                elif engine == "singularity":
                    if cty is not None and cty not in (
                            "docker",
                            "file",
                            "library",
                            "shub",
                    ):
                        raise ValueError(
                            f"singularity engine only allows docker, file, library, and shub container {cty} specified"
                        )
                elif engine is not None and engine != "local":
                    raise ValueError(
                        f"Only docker and singularity container engines are supported: {engine} specified"
                    )
                else:
                    # engine is none, need to be refered
                    if cty == "docker":
                        engine = "docker"
                    elif cty in ("file", "shub", "library"):
                        engine = "singularity"
                    elif cty == "local":
                        engine = "local"
                    else:
                        engine = "docker"
                #
                # handle different container type
                if engine == "docker":
                    from .docker.client import SoS_DockerClient

                    docker = SoS_DockerClient()
                    docker.pull(cname)
                    kwargs["engine"] = "docker"
                    kwargs["container"] = cname
                elif engine == "singularity":
                    kwargs["engine"] = "singularity"
                    from .singularity.client import SoS_SingularityClient

                    singularity = SoS_SingularityClient()
                    singularity.pull(kwargs["container"])
                else:
                    # if local or none, reset container
                    kwargs["engine"] = None
                    kwargs["container"] = None
            if "active" in kwargs:
                if kwargs["active"] is False:
                    return None
                elif kwargs["active"] is True:
                    pass
                elif isinstance(kwargs["active"], int):
                    if (kwargs["active"] >= 0 and
                            env.sos_dict["_index"] != kwargs["active"]):
                        return None
                    if (kwargs["active"] < 0 and env.sos_dict["_index"] !=
                            kwargs["active"] + env.sos_dict["__num_groups__"]):
                        return None
                elif isinstance(kwargs["active"], Sequence):
                    allowed_index = list([
                        x if x >= 0 else env.sos_dict["__num_groups__"] + x
                        for x in kwargs["active"]
                    ])
                    if env.sos_dict["_index"] not in allowed_index:
                        return None
                elif isinstance(kwargs["active"], slice):
                    allowed_index = list(range(
                        env.sos_dict["__num_groups__"]))[kwargs["active"]]
                    if env.sos_dict["_index"] not in allowed_index:
                        return None
                else:
                    raise RuntimeError(
                        f'Unacceptable value for option active: {kwargs["active"]}'
                    )
            # verify input
            if "input" in kwargs and kwargs["input"] is not None:
                try:
                    ifiles = sos_targets(kwargs["input"])
                    for ifile in ifiles:
                        if not ifile.target_exists("target"):
                            raise RuntimeError(
                                f"Input file {ifile} does not exist.")
                except Exception as e:
                    raise ValueError(
                        f'Unacceptable value ({kwargs["input"]}) for parameter input of actions: {e}'
                    )

            # if there are parameters input and output, the action is subject to signature verification
            sig = None
            # tracked can be True, filename or list of filename
            if ("tracked" in kwargs and kwargs["tracked"] is not None and
                    kwargs["tracked"] is not False):
                if args and isinstance(args[0], str):
                    script = args[0]
                elif "script" in kwargs:
                    script = kwargs["script"]
                else:
                    script = ""

                try:
                    tfiles = sos_targets(
                        [] if kwargs["tracked"] is True else kwargs["tracked"])
                except Exception as e:
                    raise ValueError(
                        f'Parameter tracked of actions can be None, True/False, or one or more filenames: {kwargs["tracked"]} provided: {e}'
                    )

                # append input and output
                for t in ("input", "output"):
                    if t in kwargs and kwargs[t] is not None:
                        tfiles.extend(sos_targets(kwargs[t]))

                from .targets import RuntimeInfo

                sig = RuntimeInfo(
                    textMD5(script),
                    sos_targets(kwargs["input"] if "input" in kwargs else []),
                    sos_targets(kwargs["output"] if "output" in kwargs else []),
                    sos_targets(kwargs["tracked"] if "tracked" in kwargs and
                                kwargs["tracked"] is not True else []),
                    kwargs,
                )
                sig.lock()
                if env.config["sig_mode"] in ("default", "skip", "distributed"):
                    matched = sig.validate()
                    if isinstance(matched, dict):
                        env.logger.info(
                            f"Action ``{func.__name__}`` is ``ignored`` due to saved signature"
                        )
                        return None
                    else:
                        env.logger.debug(f"Signature mismatch: {matched}")
                elif env.config["sig_mode"] == "assert":
                    matched = sig.validate()
                    if isinstance(matched, str):
                        raise RuntimeError(f"Signature mismatch: {matched}")
                    else:
                        env.logger.info(
                            f"Action ``{func.__name__}`` is ``ignored`` with matching signature"
                        )
                        return None
                elif env.config["sig_mode"] == "build":
                    # build signature require existence of files
                    if sig.write():
                        env.logger.info(
                            f"Action ``{func.__name__}`` is ``ignored`` with signature constructed"
                        )
                        return None
            original_env = {}
            if "default_env" in kwargs:
                original_env = copy.deepcopy(os.environ)
                if not isinstance(kwargs["default_env"], dict):
                    raise ValueError(
                        f'Option default_env must be a dictionary, {kwargs["default_env"]} provided'
                    )
                for k in kwargs["default_env"]:
                    if k not in os.environ:
                        os.environ[k] = kwargs["default_env"][k]
            if "env" in kwargs:
                original_env = copy.deepcopy(os.environ)
                if not isinstance(kwargs["env"], dict):
                    raise ValueError(
                        f'Option env must be a dictionary, {kwargs["env"]} provided'
                    )
                os.environ.update(kwargs["env"])
            # workdir refers to directory inside of docker image
            if "workdir" in kwargs:
                if not kwargs["workdir"] or not isinstance(
                        kwargs["workdir"], (str, os.PathLike)):
                    raise RuntimeError(
                        f'workdir option should be a path of type str or path, {kwargs["workdir"]} provided'
                    )
                workdir = path(kwargs["workdir"])
                if not os.path.isdir(workdir):
                    os.makedirs(workdir, exist_ok=True)
                try:
                    olddir = os.getcwd()
                    os.chdir(workdir)
                    try:
                        res = func(*args, **kwargs)
                    except Exception as e:
                        if "allow_error" in kwargs and kwargs["allow_error"]:
                            env.logger.warning(str(e))
                            res = None
                        else:
                            raise
                finally:
                    os.chdir(olddir)
                    if original_env:
                        os.environ.clear()
                        os.environ.update(original_env)
            else:
                try:
                    res = func(*args, **kwargs)
                except Exception as e:
                    if "allow_error" in kwargs and kwargs["allow_error"]:
                        env.logger.warning(str(e))
                        res = None
                    else:
                        raise
                finally:
                    if original_env:
                        os.environ.clear()
                        os.environ.update(original_env)
            if "output" in kwargs and kwargs["output"] is not None:
                ofiles = sos_targets(kwargs["output"])
                for ofile in ofiles:
                    if not ofile.target_exists("any"):
                        raise RuntimeError(
                            f"Output target {ofile} does not exist after completion of action {func.__name__}"
                        )
            if sig:
                sig.write()
                sig.release()
            return res

        return action_wrapper

    return runtime_decorator


class SoS_ExecuteScript:

    def __init__(self, script, interpreter, suffix, args=""):
        self.script = script
        self.interpreter = interpreter
        self.args = args
        if suffix:
            self.suffix = suffix
        elif sys.platform == "win32":
            self.suffix = ".bat"
        else:
            self.suffix = ".sh"

    def process_template(self, cmd, filename, script, **kwargs):
        if "template" in kwargs:
            template = kwargs["template"]
        else:
            template_name = kwargs["template_name"]
            if "CONFIG" not in env.sos_dict:
                load_config_files()
            if ("action_templates" in env.sos_dict["CONFIG"] and template_name
                    in env.sos_dict["CONFIG"]["action_templates"]):
                template = env.sos_dict["CONFIG"]["action_templates"][
                    template_name]
            elif template_name == "conda":
                template = textwrap.dedent("""\
                    conda run -n {env_name} {cmd}
                    """)
            else:
                raise ValueError(
                    f'No template named {template_name} is built-in or provided in "action_templates" of config files.'
                )

        try:
            context = copy.deepcopy(kwargs)
            context["cmd"] = cmd
            context["filename"] = filename
            context["script"] = script
            return interpolate(template, context)
        except Exception as e:
            raise ValueError(f"Failed to expand template {template}: {e}")

    def run(self, **kwargs):
        #
        if "input" in kwargs:
            try:
                ifiles = sos_targets(kwargs["input"])
            except Exception as e:
                raise ValueError(
                    f'Unacceptable value ({kwargs["input"]}) for paremter input: {e}'
                )

            content = ""
            for ifile in ifiles:
                try:
                    with open(ifile) as iscript:
                        content += iscript.read()
                except Exception as e:
                    raise RuntimeError(f"Failed to read from {ifile}: {e}")
            self.script = content + self.script

        if "engine" in kwargs and kwargs["engine"] == "docker":
            from .docker.client import SoS_DockerClient

            docker = SoS_DockerClient()
            docker.run(
                kwargs["container"],
                self.script,
                self.interpreter,
                self.args,
                self.suffix,
                **kwargs,
            )
        elif "engine" in kwargs and kwargs["engine"] == "singularity":
            from .singularity.client import SoS_SingularityClient

            singularity = SoS_SingularityClient()
            singularity.run(
                kwargs["container"],
                self.script,
                self.interpreter,
                self.args,
                self.suffix,
                **kwargs,
            )
        else:
            if isinstance(self.interpreter, str):
                if self.interpreter and not shutil.which(
                        shlex.split(self.interpreter)[0]):
                    raise RuntimeError(
                        f"Failed to locate interpreter {self.interpreter}")
            elif isinstance(self.interpreter, Sequence):
                found = False
                for ip in self.interpreter:
                    if shutil.which(shlex.split(ip)[0]):
                        self.interpreter = ip
                        found = True
                        break
                if not found:
                    raise RuntimeError(
                        f'Failed to locate any of the interpreters {", ".join(self.interpreter)}'
                    )
            else:
                raise RuntimeError(
                    f"Unacceptable interpreter {self.interpreter}")

            debug_script_file = os.path.join(
                env.exec_dir,
                f'{env.sos_dict["step_name"]}_{env.sos_dict["_index"]}_{str(uuid.uuid4())[:8]}{self.suffix}',
            )
            # with open(debug_script_file, 'w') as sfile:
            #    sfile.write(self.script)
            # env.log_to_file('ACTION', self.script)

            try:
                p = None
                script_file = tempfile.NamedTemporaryFile(
                    mode="w+t", suffix=self.suffix, delete=False).name
                # potentially used for template
                cmd_file = None
                with open(script_file, "w") as sfile:
                    sfile.write(self.script)
                if not self.args:
                    self.args = "{filename:q}"
                # if no intepreter, let us prepare for the case when the script will be executed directly
                if not self.interpreter:
                    # make the script executable
                    os.chmod(script_file, 0o775)
                #
                if env.config["run_mode"] == "dryrun":
                    cmd = interpolate(
                        f"{self.interpreter} {self.args}",
                        {
                            "filename": path("SCRIPT"),
                            "script": self.script
                        },
                    )
                    if "__std_out__" in env.sos_dict:
                        with open(env.sos_dict["__std_out__"], "a") as so:
                            so.write(f"HINT: {cmd}\n{self.script}\n")
                    else:
                        print(f"HINT: {cmd}\n{self.script}\n")
                    return None
                cmd = interpolate(
                    f"{self.interpreter} {self.args}",
                    {
                        "filename": sos_targets(script_file),
                        "script": self.script
                    },
                )
                transcript_cmd = interpolate(
                    f"{self.interpreter} {self.args}",
                    {
                        "filename": sos_targets("SCRIPT"),
                        "script": self.script
                    },
                )
                if "template_name" in kwargs or "template" in kwargs:
                    templated_script = self.process_template(
                        cmd, sos_targets(script_file), self.script, **kwargs)
                    cmd_file = tempfile.NamedTemporaryFile(
                        mode="w+t",
                        suffix=".bat" if sys.platform == "win32" else ".sh",
                        delete=False,
                    ).name
                    with open(cmd_file, "w") as cfile:
                        cfile.write(templated_script)
                    # if it has an shebang line
                    if templated_script.startswith(
                            "#!") or sys.platform == "win32":
                        os.chmod(cmd_file, 0o775)
                        cmd = cmd_file
                    else:
                        cmd = f"sh {shlex.quote(cmd_file)}"
                    env.logger.debug(
                        f"Running templated script \n{templated_script}\ncommand {cmd}"
                    )
                transcribe(self.script, cmd=transcript_cmd)
                # if not notebook, not task, signature database is avaialble.
                if (env.sos_dict["_index"] == 0 and
                        env.config["run_mode"] != "interactive" and
                        "__std_out__" not in env.sos_dict and
                        hasattr(env, "master_push_socket") and
                        env.master_push_socket is not None):
                    send_message_to_controller([
                        "workflow_sig",
                        "transcript",
                        env.sos_dict["step_name"],
                        repr({
                            "start_time": time.time(),
                            "command": transcript_cmd,
                            "script": self.script,
                        }),
                    ])

                if env.config["run_mode"] == "interactive":
                    if "stdout" in kwargs or "stderr" in kwargs:
                        child = subprocess.Popen(
                            cmd,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            bufsize=0,
                        )
                        out, err = child.communicate()
                        if "stdout" in kwargs:
                            if kwargs["stdout"] is not False and len(out):
                                with open(kwargs["stdout"], "ab") as so:
                                    so.write(out)
                        else:
                            sys.stdout.write(out.decode())

                        if "stderr" in kwargs:
                            if kwargs["stderr"] is not False and len(err):
                                with open(kwargs["stderr"], "ab") as se:
                                    se.write(err)
                        else:
                            sys.stderr.write(err.decode())
                        ret = child.returncode
                    else:
                        # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
                        from .utils import pexpect_run

                        ret = pexpect_run(cmd.strip())
                elif "__std_out__" in env.sos_dict and "__std_err__" in env.sos_dict:
                    if "stdout" in kwargs or "stderr" in kwargs:
                        if "stdout" in kwargs:
                            if kwargs["stdout"] is False:
                                so = subprocess.DEVNULL
                            else:
                                so = open(kwargs["stdout"], "ab")
                        elif env.verbosity > 0:
                            so = open(env.sos_dict["__std_out__"], "ab")
                        else:
                            so = subprocess.DEVNULL

                        if "stderr" in kwargs:
                            if kwargs["stderr"] is False:
                                se = subprocess.DEVNULL
                            else:
                                se = open(kwargs["stderr"], "ab")
                        elif env.verbosity > 1:
                            se = open(env.sos_dict["__std_err__"], "ab")
                        else:
                            se = subprocess.DEVNULL

                        p = subprocess.Popen(
                            cmd, shell=True, stderr=se, stdout=so)
                        ret = p.wait()

                        if so != subprocess.DEVNULL:
                            so.close()
                        if se != subprocess.DEVNULL:
                            se.close()

                    elif env.verbosity >= 1:
                        with open(env.sos_dict["__std_out__"],
                                  "ab") as so, open(env.sos_dict["__std_err__"],
                                                    "ab") as se:
                            p = subprocess.Popen(
                                cmd, shell=True, stderr=se, stdout=so)
                            ret = p.wait()
                    else:
                        p = subprocess.Popen(
                            cmd,
                            shell=True,
                            stderr=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                        )
                        ret = p.wait()
                else:
                    if "stdout" in kwargs:
                        if kwargs["stdout"] is False:
                            so = subprocess.DEVNULL
                        else:
                            so = open(kwargs["stdout"], "ab")
                    elif env.verbosity > 0:
                        so = None
                    else:
                        so = subprocess.DEVNULL

                    if "stderr" in kwargs:
                        if kwargs["stderr"] is False:
                            se = subprocess.DEVNULL
                        else:
                            se = open(kwargs["stderr"], "ab")
                    elif env.verbosity > 1:
                        se = None
                    else:
                        se = subprocess.DEVNULL

                    p = subprocess.Popen(cmd, shell=True, stderr=se, stdout=so)

                    ret = p.wait()
                    if so is not None and so != subprocess.DEVNULL:
                        so.close()
                    if se is not None and se != subprocess.DEVNULL:
                        se.close()
                # clean up empty stdstream files
                for item in ["stdout", "stderr"]:
                    if (item in kwargs and os.path.isfile(kwargs[item]) and
                            os.path.getsize(kwargs[item]) == 0):
                        try:
                            os.remove(kwargs[item])
                        except Exception:
                            pass
                if ret != 0:
                    with open(debug_script_file, "w") as sfile:
                        sfile.write(self.script)
                    cmd = cmd.replace(script_file, debug_script_file)
                    out = (f", stdout={kwargs['stdout']}"
                           if "stdout" in kwargs and
                           os.path.isfile(kwargs["stdout"]) and
                           os.path.getsize(kwargs["stdout"]) > 0 else "")
                    err = (f", stderr={kwargs['stderr']}"
                           if "stderr" in kwargs and
                           os.path.isfile(kwargs["stderr"]) and
                           os.path.getsize(kwargs["stderr"]) > 0 else "")
                    raise subprocess.CalledProcessError(
                        returncode=ret,
                        cmd=cmd,
                        stderr="\nFailed to execute ``{}``\nexitcode={}, workdir=``{}``{}{}{}\n{}"
                        .format(
                            cmd,
                            ret,
                            os.getcwd(),
                            f', task={os.path.basename(env.sos_dict["__std_err__"]).split(".")[0]}'
                            if "__std_err__" in env.sos_dict else "",
                            out,
                            err,
                            "-" * 75,
                        ),
                    )
            finally:
                try:
                    os.remove(script_file)
                except Exception:
                    # 1315: ignore in case the temp script file no longer exists
                    pass
                if cmd_file is not None:
                    try:
                        os.remove(cmd_file)
                    except Exception:
                        # 1315: ignore in case the temp script file no longer exists
                        pass


@SoS_Action()
def sos_run(workflow=None,
            targets=None,
            shared=None,
            args=None,
            source=None,
            **kwargs):
    """Execute a workflow from the current SoS script or a specified source
    (in .sos or .ipynb format), with _input as the initial input of workflow."""
    if "__std_out__" in env.sos_dict and "__std_err__" in env.sos_dict:
        raise RuntimeError(
            "Executing nested workflow (action sos_run) in tasks is not supported."
        )

    if isinstance(workflow, str):
        workflows = [workflow]
    elif isinstance(workflow, Sequence):
        workflows = list(workflow)
    elif workflow is None:
        workflows = []
    else:
        raise ValueError(
            "workflow has to be None, a workflow name, or a list of workflow names"
        )

    if source is None:
        script = SoS_Script(
            env.sos_dict["__step_context__"].content,
            env.sos_dict["__step_context__"].filename,
        )
        if workflows:
            wfs = [script.workflow(wf, use_default=True) for wf in workflows]
        else:
            wfs = [script.workflow(use_default=False)]
    else:
        # reading workflow from another file
        script = SoS_Script(filename=source)

        if workflows:
            wfs = [script.workflow(wf, use_default=True) for wf in workflows]
        else:
            wfs = [script.workflow(use_default=False)]

    # if wf contains the current step or one of the previous one, this constitute
    # recusive nested workflow and should not be allowed
    all_parameters = set()
    for wf in wfs:
        all_parameters |= set(wf.parameters())
        if env.sos_dict["step_name"] in [
                f"{x.name}_{x.index}" for x in wf.sections
        ]:
            raise RuntimeError(
                f'Nested workflow {workflow} contains the current step {env.sos_dict["step_name"]}'
            )

    # args can be specified both as a dictionary or keyword arguments
    if args is None:
        args = kwargs
    else:
        args.update(kwargs)

    for key in args.keys():
        if key not in all_parameters and key not in SOS_ACTION_OPTIONS:
            raise ValueError(
                f"No parameter {key} is defined for workflow {workflow}")

    if shared is None:
        shared = []
    elif isinstance(shared, str):
        shared = [shared]

    # for nested workflow, _input would becomes the input of workflow.
    env.sos_dict.set("__step_output__",
                     copy.deepcopy(env.sos_dict.get("_input", None)))
    shared.append("__step_output__")
    try:
        my_name = env.sos_dict["step_name"]
        args_output = ", ".join(f"{x}={short_repr(y)}" for x, y in args.items()
                                if not x.startswith("__"))
        if "ACTION" in env.config["SOS_DEBUG"] or "ALL" in env.config[
                "SOS_DEBUG"]:
            env.log_to_file(
                "ACTION",
                "Executing workflow ``{}`` with input ``{}`` and {}".format(
                    workflow,
                    short_repr(env.sos_dict.get("_input", None), True),
                    "no args" if not args_output else args_output,
                ),
            )

        if not hasattr(env, "__socket__") or env.__socket__ is None:
            raise RuntimeError(
                "sos_run function cannot be executed in scratch cell.")
        # tell the master process to receive a workflow
        # really send the workflow
        shared = {
            x: (env.sos_dict[x] if x in env.sos_dict else None) for x in shared
        }

        wf_ids = [str(uuid.uuid4()) for wf in wfs]

        blocking = not env.sos_dict.get("__concurrent_subworkflow__", False)
        env.__socket__.send(
            encode_msg([
                "workflow", wf_ids, wfs, targets, args, shared, env.config,
                blocking
            ]))

        if not blocking:
            return {"pending_workflows": wf_ids}
        res = {}
        for wf in wfs:
            wf_res = decode_msg(env.__socket__.recv())
            res.update(wf_res)
            if wf_res is None:
                sys.exit(0)
            elif isinstance(wf_res, Exception):
                raise wf_res
            else:
                env.sos_dict.quick_update(wf_res["shared"])
        return res
    finally:
        # restore step_name in case the subworkflow re-defines it
        env.sos_dict.set("step_name", my_name)


@SoS_Action(acceptable_args=["script", "interpreter", "suffix", "args"])
def script(script, interpreter="", suffix="", args="", **kwargs):
    """Execute specified script using specified interpreter. This action accepts common
    action arguments such as input, active, workdir, docker_image and args. In particular,
    content of one or more files specified by option input would be prepended before
    the specified script."""
    return SoS_ExecuteScript(script, interpreter, suffix, args).run(**kwargs)


@SoS_Action(acceptable_args=["expr", "msg"])
def fail_if(expr, msg=""):
    """Raise an exception with `msg` if condition `expr` is False"""
    if expr:
        raise TerminateExecution(
            msg if msg else "error triggered by action fail_if")
    return 0


@SoS_Action(acceptable_args=["expr", "msg"])
def warn_if(expr, msg=""):
    """Yield an warning message `msg` if `expr` is False """
    if expr:
        env.logger.warning(msg)
    return 0


@SoS_Action(acceptable_args=["expr", "msg", "no_output"])
def stop_if(expr, msg="", no_output=False):
    """Abort the execution of the current step or loop and yield
    an warning message `msg` if `expr` is False"""
    if expr:
        raise StopInputGroup(msg=msg, keep_output=not no_output)
    return 0


@SoS_Action(acceptable_args=["expr", "msg"])
def done_if(expr, msg=""):
    """Assuming that output has already been generated and stop
    executing the rest of the substep"""
    if expr:
        raise StopInputGroup(msg=msg, keep_output=True)
    return 0


@SoS_Action(acceptable_args=["expr", "msg", "no_output"])
def skip_if(expr, msg=""):
    """Skip the current substep and set _output to empty. Output
    will be removed if already generated."""
    if expr:
        raise StopInputGroup(msg=msg, keep_output=False)
    return 0


#
# download file with progress bar
#


def downloadURL(URL, dest, decompress=False, index=None):
    dest = os.path.abspath(os.path.expanduser(dest))
    dest_dir, filename = os.path.split(dest)
    #
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir, exist_ok=True)
    if not os.path.isdir(dest_dir):
        raise RuntimeError(
            f"Failed to create destination directory to download {URL}")
    #
    message = filename
    if len(message) > 30:
        message = message[:10] + "..." + message[-16:]
    #
    dest_tmp = dest + f".tmp_{os.getpid()}"
    term_width = shutil.get_terminal_size((80, 20)).columns
    try:
        env.logger.debug(f"Download {URL} to {dest}")
        sig = file_target(dest)
        if os.path.isfile(dest):
            prog = ProgressBar(
                desc=message,
                disable=env.verbosity <= 1,
                position=index,
                leave=True,
                bar_format="{desc}",
                total=10000000,
            )
            target = file_target(dest)
            if env.config["sig_mode"] == "build":
                prog.set_description(message +
                                     ": \033[32m writing signature\033[0m")
                prog.update()
                target.write_sig()
                prog.close()
                return True
            elif env.config["sig_mode"] == "ignore":
                prog.set_description(message + ": \033[32m use existing\033[0m")
                prog.update()
                prog.close()
                return True
            elif env.config["sig_mode"] in ("default", "skip", "distributed"):
                prog.update()
                if sig.validate():
                    prog.set_description(message +
                                         ": \033[32m Validated\033[0m")
                    prog.update()
                    prog.close()
                    return True
                else:
                    prog.set_description(message +
                                         ":\033[91m Signature mismatch\033[0m")
                    target.write_sig()
                    prog.update()
        #
        prog = ProgressBar(
            desc=message,
            disable=env.verbosity <= 1,
            position=index,
            leave=True,
            bar_format="{desc}",
            total=10000000,
        )
        #
        # Stop using pycurl because of libcurl version compatibility problems
        # that happen so often and difficult to fix. Error message looks like
        #
        # Reason: Incompatible library version: pycurl.cpython-35m-darwin.so
        # requires version 9.0.0 or later, but libcurl.4.dylib provides version 7.0.0
        #
        # with open(dest_tmp, 'wb') as f:
        #    c = pycurl.Curl()
        #    c.setopt(pycurl.URL, str(URL))
        #    c.setopt(pycurl.WRITEFUNCTION, f.write)
        #    c.setopt(pycurl.SSL_VERIFYPEER, False)
        #    c.setopt(pycurl.NOPROGRESS, False)
        #    c.setopt(pycurl.PROGRESSFUNCTION, prog.curlUpdate)
        #    c.perform()
        # if c.getinfo(pycurl.HTTP_CODE) == 404:
        #    prog.set_description(message + ':\033[91m 404 Error {}\033[0m'.format(' '*(term_width - len(message) - 12)))
        #    try:
        #        os.remove(dest_tmp)
        #    except OSError:
        #        pass
        #    return False
        with open(dest_tmp, "wb") as f:
            try:
                u = urllib.request.urlopen(str(URL))
                try:
                    file_size = int(u.getheader("Content-Length"))
                    prog = ProgressBar(
                        total=file_size,
                        desc=message,
                        position=index,
                        leave=False)
                except Exception:
                    file_size = None
                file_size_dl = 0
                block_sz = 8192
                while True:
                    buffer = u.read(block_sz)
                    if not buffer:
                        break
                    file_size_dl += len(buffer)
                    f.write(buffer)
                    prog.update(len(buffer))
            except urllib.error.HTTPError as e:
                prog.set_description(message +
                                     f":\033[91m {e.code} Error\033[0m")
                prog.update()
                prog.close()
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
            except Exception as e:
                prog.set_description(message + f":\033[91m {e}\033[0m")
                prog.update()
                prog.close()
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
        #
        if os.path.isfile(dest):
            os.remove(dest)
        os.rename(dest_tmp, dest)
        decompressed = 0
        if decompress:
            if zipfile.is_zipfile(dest):
                prog.set_description(message + ":\033[91m Decompressing\033[0m")
                prog.update()
                prog.close()
                zfile = zipfile.ZipFile(dest)
                zfile.extractall(dest_dir)
                names = zfile.namelist()
                for name in names:
                    if os.path.isdir(os.path.join(dest_dir, name)):
                        continue
                    elif not os.path.isfile(os.path.join(dest_dir, name)):
                        return False
                    else:
                        decompressed += 1
            elif tarfile.is_tarfile(dest):
                prog.set_description(message + ":\033[91m Decompressing\033[0m")
                prog.update()
                prog.close()
                with tarfile.open(dest, "r:*") as tar:
                    tar.extractall(dest_dir)
                    # only extract files
                    files = [x.name for x in tar.getmembers() if x.isfile()]
                    for name in files:
                        if not os.path.isfile(os.path.join(dest_dir, name)):
                            return False
                        else:
                            decompressed += 1
            elif dest.endswith(".gz"):
                prog.set_description(message + ":\033[91m Decompressing\033[0m")
                prog.update()
                prog.close()
                decomp = dest[:-3]
                with gzip.open(dest, "rb") as fin, open(decomp, "wb") as fout:
                    buffer = fin.read(100000)
                    while buffer:
                        fout.write(buffer)
                        buffer = fin.read(100000)
                decompressed += 1
        decompress_msg = (
            "" if not decompressed else
            f' ({decompressed} file{"" if decompressed <= 1 else "s"} decompressed)'
        )
        prog.set_description(
            message +
            f':\033[32m downloaded{decompress_msg} {" "*(term_width - len(message) - 13 - len(decompress_msg))}\033[0m'
        )
        prog.update()
        prog.close()
        # if a md5 file exists
        # if downloaded files contains .md5 signature, use them to validate
        # downloaded files.
        if os.path.isfile(dest + ".md5"):
            prog.set_description(message +
                                 ":\033[91m Verifying md5 signature\033[0m")
            prog.update()
            prog.close()
            with open(dest + ".md5") as md5:
                rec_md5 = md5.readline().split()[0].strip()
                obs_md5 = fileMD5(dest, partial=False)
                if rec_md5 != obs_md5:
                    prog.set_description(
                        message + ":\033[91m MD5 signature mismatch\033[0m")
                    prog.update()
                    prog.close()
                    env.logger.warning(
                        f"md5 signature mismatch for downloaded file {filename[:-4]} (recorded {rec_md5}, observed {obs_md5})"
                    )
            prog.set_description(message +
                                 ":\033[91m MD5 signature verified\033[0m")
            prog.update()
            prog.close()
    except Exception as e:
        if env.verbosity > 2:
            sys.stderr.write(get_traceback())
        env.logger.error(f"Failed to download: {e}")
        return False
    finally:
        # if there is something wrong still remove temporary file
        if os.path.isfile(dest_tmp):
            os.remove(dest_tmp)
    return os.path.isfile(dest)


@SoS_Action(acceptable_args=[
    "URLs",
    "workdir",
    "dest_dir",
    "dest_file",
    "decompress",
    "max_jobs",
])
def download(URLs, dest_dir=".", dest_file=None, decompress=False, max_jobs=5):
    """Download files from specified URL, which should be space, tab or
    newline separated URLs. The files will be downloaded to specified destination.
    Option "dest_dir" specify the destination directory,
    and "dest_file" specify the output filename, which will otherwise be the same
    specified in the URL. If `filename.md5` files are downloaded, they are used to
    validate downloaded `filename`. If "decompress=True", compressed
    files are decompressed. If `max_jobs` is given, a maximum of `max_jobs`
    concurrent download jobs will be used for each domain. This restriction
    applies to domain names and will be applied to multiple download
    instances.
    """
    if env.config["run_mode"] == "dryrun":
        print(f"HINT: download\n{URLs}\n")
        return None
    if isinstance(URLs, str):
        urls = [x.strip() for x in URLs.split() if x.strip()]
    else:
        urls = list(URLs)

    if not urls:
        env.logger.debug(f"No download URL specified: {URLs}")
        return
    #
    if dest_file is not None and len(urls) != 1:
        raise RuntimeError(
            "Only one URL is allowed if a destination file is specified.")
    #
    if dest_file is None:
        filenames = []
        for idx, url in enumerate(urls):
            token = urllib.parse.urlparse(url)
            # if no scheme or netloc, the URL is not acceptable
            if not all([
                    getattr(token, qualifying_attr)
                    for qualifying_attr in ("scheme", "netloc")
            ]):
                raise ValueError(f"Invalid URL {url}")
            filename = os.path.split(token.path)[-1]
            if not filename:
                raise ValueError(f"Cannot determine destination file for {url}")
            filenames.append(os.path.join(dest_dir, filename))
    else:
        token = urllib.parse.urlparse(urls[0])
        if not all([
                getattr(token, qualifying_attr)
                for qualifying_attr in ("scheme", "netloc")
        ]):
            raise ValueError(f"Invalid URL {url}")
        filenames = [dest_file]
    #
    succ = [(False, None) for x in urls]
    with ProcessPoolExecutor(max_workers=max_jobs) as executor:
        for idx, (url, filename) in enumerate(zip(urls, filenames)):
            # if there is alot, start download
            succ[idx] = executor.submit(downloadURL, url, filename, decompress,
                                        idx)
    succ = [x.result() for x in succ]

    # for su, url in zip(succ, urls):
    #    if not su:
    #        env.logger.warning('Failed to download {}'.format(url))
    failed = [y for x, y in zip(succ, urls) if not x]
    if failed:
        if len(urls) == 1:
            raise RuntimeError("Failed to download {urls[0]}")
        else:
            raise RuntimeError(
                f"Failed to download {failed[0]} ({len(failed)} out of {len(urls)})"
            )
    return 0


@SoS_Action(acceptable_args=["script", "args"])
def run(script, args="", **kwargs):
    """Execute specified script using bash. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    if sys.platform == "win32":
        # in the case there is no interpreter, we put the script
        # at first (this is the case for windows)
        # and we donot add default args.
        interpreter = ""
    else:
        # if there is a shebang line, we ...
        if not script.startswith("#!"):
            interpreter = "/bin/bash"
            if not args:
                args = "-ev {filename:q}"
        else:
            # execute script directly
            interpreter = ""
    return SoS_ExecuteScript(script, interpreter, "", args).run(**kwargs)


@SoS_Action(acceptable_args=["script", "args"])
def perl(script, args="", **kwargs):
    """Execute specified script using perl. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, "perl", ".pl", args).run(**kwargs)


def collect_input(script, input):
    # determine file extension
    if input is not None:
        if isinstance(input, (str, file_target)):
            ext = os.path.splitext(input)[-1]
        elif isinstance(input, Sequence) and len(input) > 0:
            ext = os.path.splitext(input[0])[-1]
        else:
            raise ValueError("Unknown input file for action pandoc")
    else:
        ext = ".md"

    input_file = tempfile.NamedTemporaryFile(
        mode="w+t", suffix=ext, delete=False).name
    with open(input_file, "w") as tmp:
        if script is not None and script.strip():
            tmp.write(script.rstrip() + "\n\n")
        if isinstance(input, str):
            try:
                with open(input) as ifile:
                    tmp.write(ifile.read() + "\n\n")
            except Exception as e:
                raise ValueError(f"Failed to read input file {input}: {e}")
        elif isinstance(input, Sequence):
            for ifile in input:
                try:
                    with open(ifile) as itmp:
                        tmp.write(itmp.read().rstrip() + "\n\n")
                except Exception as e:
                    raise ValueError(f"Failed to read input file {ifile}: {e}")
    return input_file


@SoS_Action(acceptable_args=["script"])
def report(script=None, input=None, output=None, **kwargs):
    """Write script to an output file specified by `output`, which can be
    a filename to which the content of the script will be written,
    any object with a "write" attribute (e.g. a file handle) for which the "write"
    function will be called with the content. If output is unspecified, the content
    will be written to standard output or appended to a file specified with command
    line option `-r`."""
    if env.config["run_mode"] == "dryrun":
        if "__std_out__" in env.sos_dict:
            with open(env.sos_dict["__std_out__"], "a") as so:
                so.write(f'HINT: report:\n{"" if script is None else script}\n')
                if input is not None:
                    for ifile in input:
                        so.write(f"  from file: {ifile}\n")
        else:
            print(f'HINT: report:\n{"" if script is None else script}')
            if input is not None:
                for ifile in input:
                    print(f"  from file: {ifile}")
        return
    file_handle = None
    if isinstance(output, str):
        if not output or output == "-":
            writer = sys.stdout.write
        elif output.startswith(">>"):
            file_handle = open(os.path.expanduser(output[2:]), "a")
            writer = file_handle.write
        else:
            file_handle = open(os.path.expanduser(output), "w")
            writer = file_handle.write
    elif isinstance(output, (path, file_target)):
        file_handle = open(os.path.expanduser(str(output)), "w")
        writer = file_handle.write
    elif isinstance(output, (paths, sos_targets)):
        if len(output) != 1:
            raise ValueError(f"More than one output is specified {output}")
        if not isinstance(output[0], (file_target, path)):
            raise ValueError(
                "Action report can only output to file target or standard output"
            )
        file_handle = open(os.path.expanduser(str(output[0])), "w")
        writer = file_handle.write
    elif hasattr(output, "write"):
        writer = output.write
    elif output is None or output == "":
        writer = sys.stdout.write
    else:
        raise ValueError(f"Invalid output {output}.")

    # file lock to prevent race condition
    with TimeoutInterProcessLock(os.path.join(env.temp_dir, "report_lock")):
        if isinstance(script, str) and script.strip():
            writer(script.rstrip() + "\n\n")
        if input is not None:
            if isinstance(input, (str, file_target)):
                if ("ACTION" in env.config["SOS_DEBUG"] or
                        "ALL" in env.config["SOS_DEBUG"]):
                    env.log_to_file("ACTION", f"Loading report from {input}")
                with open(input) as ifile:
                    writer(ifile.read().rstrip() + "\n\n")
            elif isinstance(input, Sequence):
                for ifile in input:
                    try:
                        env.logger.debug(f"Loading report from {ifile}")
                        with open(ifile) as itmp:
                            writer(itmp.read().rstrip() + "\n\n")
                    except Exception as e:
                        raise ValueError(
                            f"Failed to read input file {ifile}: {e}")
            else:
                raise ValueError("Unknown input file for action report")
    #
    if file_handle:
        file_handle.close()


@SoS_Action(acceptable_args=["script", "args"])
def pandoc(script=None,
           input=None,
           output=None,
           args="{input:q} --output {output:q}",
           **kwargs):
    """Convert input file to output using pandoc

    The input can be specified in three ways:

    1. instant script, which is assumed to be in md format

    pandoc:   output='report.html'
      script

    2. one or more input files. The format is determined by extension of input file

    pandoc(input, output='report.html')

    3. input file specified by command line option `-r` .
    pandoc(output='report.html')

    If no output is specified, it is assumed to be in html format
    and is written to standard output.

    You can specify more options such as "from" and "to" by customizing
    the args parameter of the action. The default value of args is
    `{input:q} --output {output:q}'
    """
    #
    #     # this is output format
    #     pandoc [OPTIONS] [FILES]
    # Input formats:  commonmark, docbook, docx, epub, haddock, html, json*, latex,
    #                 markdown, markdown_github, markdown_mmd, markdown_phpextra,
    #                 markdown_strict, mediawiki, native, odt, opml, org, rst, t2t,
    #                 textile, twiki
    #                 [ *only Pandoc's JSON version of native AST]
    # Output formats: asciidoc, beamer, commonmark, context, docbook, docx, dokuwiki,
    #                 dzslides, epub, epub3, fb2, haddock, html, html5, icml, json*,
    #                 latex, man, markdown, markdown_github, markdown_mmd,
    #                 markdown_phpextra, markdown_strict, mediawiki, native, odt,
    #                 opendocument, opml, org, pdf**, plain, revealjs, rst, rtf, s5,
    #                 slideous, slidy, tei, texinfo, textile
    #                 [**for pdf output, use latex or beamer and -o FILENAME.pdf]
    # Options:
    #   -f FORMAT, -r FORMAT  --from=FORMAT, --read=FORMAT
    #   -t FORMAT, -w FORMAT  --to=FORMAT, --write=FORMAT
    #   -o FILENAME           --output=FILENAME
    #                         --data-dir=DIRECTORY
    #   -R                    --parse-raw
    #   -S                    --smart
    #
    # IGNORED
    #
    if not executable("pandoc").target_exists():
        raise RuntimeError("pandoc not found")

    input = sos_targets(collect_input(script, input))

    output = sos_targets(output)
    if len(output) == 0:
        write_to_stdout = True
        output = sos_targets(
            tempfile.NamedTemporaryFile(
                mode="w+t", suffix=".html", delete=False).name)
    else:
        write_to_stdout = False
    #
    ret = 1
    try:
        p = None
        cmd = interpolate(f"pandoc {args}", {"input": input, "output": output})
        if "ACTION" in env.config["SOS_DEBUG"] or "ALL" in env.config[
                "SOS_DEBUG"]:
            env.log_to_file("ACTION", f'Running command "{cmd}"')
        if env.config["run_mode"] == "interactive":
            # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
            from .utils import pexpect_run

            ret = pexpect_run(cmd)
        else:
            p = subprocess.Popen(cmd, shell=True)
            ret = p.wait()
    except Exception as e:
        env.logger.error(str(e))
    if ret != 0:
        temp_file = os.path.join(".sos", f"pandoc_{os.getpid()}.md")
        shutil.copyfile(input, temp_file)
        cmd = interpolate(
            f"pandoc {args}",
            {
                "input": sos_targets(temp_file),
                "output": sos_targets(output)
            },
        )
        raise RuntimeError(
            f"Failed to execute script. Please use command \n{cmd}\nunder {os.getcwd()} to test it."
        )
    if write_to_stdout:
        with open(output[0].fullname()) as out:
            sys.stdout.write(out.read())
    else:
        env.logger.info(f"Report saved to {output}")
    try:
        os.remove(input)
    except Exception:
        pass



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import base64
import copy
import glob
import multiprocessing as mp
import os
import shutil
import socket
import stat
import subprocess
import sys
from collections.abc import Sequence
from typing import Any, Dict, List, Optional, Union

import pexpect
import pkg_resources

from .eval import Undetermined, cfg_interpolate, get_config
from .syntax import SOS_LOGLINE
from .targets import path, sos_targets
from .task_engines import BackgroundProcess_TaskEngine
from .tasks import TaskFile
from .utils import (env, expand_size, expand_time, format_HHMMSS, short_repr,
                    textMD5)
from .workflow_engines import BackgroundProcess_WorkflowEngine

#
# A 'queue' is defined by queue configurations in SoS configuration files.
# It encapsulate properties of a queue and tells sos how to interact with
# the queue. A queue can be a local host, a remote host without queue, or
# a remote host with a task queue, or a RQ or Celery server. There are
# two main categories of properties of a host.
#
# 1. host properties, namely how to copy files and how to execute comamnds
#   on the host. Note that even for queues that communicate with sos
#   through network, the workers might be on a different host with different
#   file systems.
#
#   Keys for host configuration include:
#   * path_map: path map between local and remote hosts
#   * shared: paths that are shared between local and remote hosts
#   * send_cmd (optional): alternative command to send files
#   * receive_cmd (optional): alternative command to receive files
#   * execute_cmd (optional): alternative command to execute commands
#
# 2. task properties, namely how to manage running jobs. These include
#   direct execution, PBS and various cluster systems, and various task
#   queues.
#
#   Keys for task configuration depend largely one queue type.
#
#   * task_engine: type of task engine
#   * max_jobs: maximum number of concurrent jobs on the host.
#
#
# Implementation wise, a queue instance is created for each queue.
#


class DaemonizedProcess(mp.Process):

    def __init__(self, cmd, *args, **kwargs):
        super(DaemonizedProcess, self).__init__(*args, **kwargs)
        self.cmd = cmd

    def run(self):
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            env.logger.error(f"_Fork #1 failed: {err}\n")
            sys.exit(1)

        os.setsid()
        os.umask(0)
        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            env.logger.error(f"_Fork #2 failed: {err}\n")
            sys.exit(1)
        # the following is also need to properly daemonize the process
        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        try:
            si = open(os.devnull, "r")
            so = open(os.devnull, "w")
            se = open(os.devnull, "w")
            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())
        except Exception:
            # #493
            pass

        # fork a new process
        subprocess.Popen(self.cmd, shell=True, close_fds=True)
        return


def _show_err_and_out(task_id, res) -> None:
    if "stdout" in res:
        sys.stderr.write(f"\n{task_id}.out:\n")
        ends_with_newline = False
        for line in res["stdout"].splitlines():
            if not SOS_LOGLINE.match(line):
                sys.stderr.write(line)
                ends_with_newline = line.endswith("\n")
        if not ends_with_newline:
            sys.stderr.write("\n")
    if "stderr" in res:
        sys.stderr.write(f"\n{task_id}.err:\n")
        ends_with_newline = False
        for line in res["stderr"].splitlines():
            if not SOS_LOGLINE.match(line):
                sys.stderr.write(line)
                ends_with_newline = line.endswith("\n")
        if not ends_with_newline:
            sys.stderr.write("\n")


class LocalHost(object):
    """For local host, no path map, send and receive ..."""

    def __init__(
        self,
        config: Dict[str, Union[str, int, List[str]]],
        test_connection: bool = True,
    ) -> None:
        super(LocalHost, self).__init__()

        # even if the config has an alias, we use localhost to make it clear that the host is localhost
        self.alias = config.get("alias", "localhost")
        self.address = "localhost"
        # we checkk local jobs more aggressively
        self.config = {"alias": self.alias, "status_check_interval": 2}
        self.config.update(config)

    def test_connection(self):
        return "OK"

    def target_exists(self, targets):
        return targets.target_exists()

    def target_signature(self, targets):
        return targets.target_signature()

    def send_to_host(self, items):
        return {x: x for x in items}

    def receive_from_host(self, items):
        return {x: x for x in items}

    def prepare_task(self, task_id):
        task_file = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".task")
        # add server restriction on task file
        if not os.path.isfile(task_file):
            raise ValueError(f"Missing task definition {task_file}")

        tf = TaskFile(task_id)
        params, old_runtime = tf.get_params_and_runtime()
        # clear possible previous result
        task_vars = params.sos_dict
        runtime = {
            "_runtime": {
                x: task_vars["_runtime"][x] for x in (
                    "verbosity",
                    "sig_mode",
                    "run_mode",
                    "walltime",
                    "cores",
                    "mem",
                ) if x in task_vars["_runtime"]
            }
        }
        runtime["_runtime"]["workdir"] = (
            task_vars["_runtime"]["workdir"]
            if "workdir" in task_vars["_runtime"] else os.getcwd())

        if ("max_mem" in self.config or "max_cores" in self.config or
                "max_walltime" in self.config):
            for key in ("max_mem", "max_cores", "max_walltime"):
                if key in self.config:
                    runtime["_runtime"][key] = (
                        format_HHMMSS(self.config[key])
                        if key == "max_walltime" else self.config[key])

            if (self.config.get("max_mem", None) is not None and
                    task_vars["_runtime"].get("mem", None) is not None and
                    self.config["max_mem"] < task_vars["_runtime"]["mem"]):
                env.logger.error(
                    f'Task {task_id} requested more mem ({task_vars["_runtime"]["mem"]}) than allowed max_mem ({self.config["max_mem"]})'
                )
                return False
            if (self.config.get("max_cores", None) is not None and
                    task_vars["_runtime"].get("cores", None) is not None and
                    self.config["max_cores"] < task_vars["_runtime"]["cores"]):
                env.logger.error(
                    f'Task {task_id} requested more cores ({task_vars["_runtime"]["cores"]}) than allowed max_cores ({self.config["max_cores"]})'
                )
                return False
            if (self.config.get("max_walltime", None) is not None and
                    task_vars["_runtime"].get("walltime", None) is not None and
                    expand_time(self.config["max_walltime"]) < expand_time(
                        task_vars["_runtime"]["walltime"])):
                env.logger.error(
                    f'Task {task_id} requested more walltime ({task_vars["_runtime"]["walltime"]}) than allowed max_walltime ({self.config["max_walltime"]})'
                )
                return False

        # if the task has been running remotely, we need to reset runtime for local execution
        if len(runtime) > 1 or runtime["_runtime"] or runtime != old_runtime:
            tf.runtime = runtime
        tf.status = "pending"
        self.send_job_file(task_file)
        return True

    def send_job_file(self, job_file, dir="tasks"):
        # on the same file system, no action is needed.
        dest_job_file = path(f"~/.sos/{dir}/{os.path.basename(job_file)}")
        job_file = path(job_file)
        if job_file != dest_job_file:
            shutil.copyfile(job_file, dest_job_file)

    def check_output(self, cmd, under_workdir=False, **kwargs):
        # get the output of command
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(cmd)
            return subprocess.check_output(
                cmd, shell=isinstance(cmd, str), **kwargs).decode()
        except Exception as e:
            env.logger.warning(f"Check output of {cmd} failed: {e}")
            raise

    def check_call(self, cmd, under_workdir=False, **kwargs):
        # get the output of command
        try:
            return subprocess.check_call(
                cmd, shell=isinstance(cmd, str), **kwargs)
        except Exception as e:
            env.logger.warning(f"Check output of {cmd} failed: {e}")
            raise

    def run_command(self, cmd, wait_for_task, realtime=False, **kwargs):
        # run command but does not wait for result.
        if realtime:
            from .utils import pexpect_run

            return pexpect_run(cmd)
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        if wait_for_task or sys.platform == "win32":
            p = subprocess.Popen(cmd, shell=True, **kwargs)
            p.wait()
        else:
            p = DaemonizedProcess(cmd, **kwargs)
            p.start()
            p.join()

    def receive_result(self, task_id: str) -> Dict[str, Any]:
        tf = TaskFile(task_id)

        res = tf.result
        if not res or "ret_code" not in res:
            return {
                "ret_code":
                    1,
                "task":
                    task_id,
                "exception":
                    ValueError(f"No result is received for task {task_id}"),
            }

        try:
            if res["ret_code"] != 0 or env.verbosity >= 3:
                _show_err_and_out(task_id, res)
        except Exception as e:
            # if ret_code does not exist...
            return {"ret_code": 1, "output": {}, "exception": e}
        return res


class RemoteHost(object):
    """A remote host class that manages how to communicate with remote host"""

    def __init__(
        self,
        config: Dict[str, Union[str, int, List[str]]],
        test_connection: bool = True,
    ) -> None:
        self.config = config
        self.cm_opts = self._get_control_master_options()
        self.pem_opts = self._get_identify_file_options()
        self.alias = self.config["alias"]
        #
        self.address = self.config["address"]
        self.port = self.config.get("port", 22)
        self.pem_file = self.config.get("pem_file", None)
        self.shared_dirs = self._get_shared_dirs()
        self.path_map = self._get_path_map()
        # we already test connect of remote hosts
        if test_connection:
            test_res = self.test_connection()
            if test_res != "OK":
                raise RuntimeError(
                    f"Failed to connect to {self.alias}: {test_res}")

    def target_exists(self, targets):
        try:
            msg = self.check_output(
                [
                    "sos",
                    "preview",
                    "--exists",
                    base64.b64encode(repr(targets).encode()).decode(),
                ],
                under_workdir=True,
            ).strip()
        except Exception as e:
            msg = f"error: {e}"
        if msg.startswith("error:"):
            env.logger.debug(msg)
            return True
        return msg == "yes"

    def target_signature(self, targets):
        try:
            msg = self.check_output(
                [
                    "sos",
                    "preview",
                    "--signature",
                    base64.b64encode(repr(targets).encode()).decode(),
                ],
                under_workdir=True,
            ).strip()
        except Exception as e:
            msg = f"error: {e}"
        if msg.startswith("error:"):
            env.logger.debug(msg)
            return textMD5(targets.target_name())
        else:
            return msg

    def _get_shared_dirs(self) -> List[Any]:
        value = self.config.get("shared", [])
        if isinstance(value, str):
            return [value]
        if isinstance(value, Sequence):
            return value
        raise ValueError(
            "Option shared can only be a string or a list of strings")

    def _get_path_map(self) -> Dict[str, str]:
        res: Dict = {}
        # if user-specified path_map, it overrides CONFIG
        path_map = self.config.get("path_map", [])
        #
        if not path_map:
            return res
        if isinstance(path_map, str):
            path_map = [path_map]
        if isinstance(path_map, Sequence):
            for v in path_map:
                if " -> " not in v:
                    raise ValueError(
                        f"Path map should be separated as from -> to, {v} specified"
                    )
                elif v.count(" -> ") > 1:
                    raise ValueError(
                        f"Path map should be separated as from -> to, {v} specified"
                    )
                res[v.split(" -> ")[0]] = v.split(" -> ")[1]
        elif isinstance(path_map, dict):
            for k, v in path_map.items():
                res[k] = v
        else:
            raise ValueError(
                f"Unacceptable path_mapue for configuration path_map: {path_map}"
            )
        return res

    def _get_control_master_options(self):
        master_dir = os.path.join(
            os.path.expanduser("~"), ".ssh", "controlmasters")
        if not os.path.isdir(master_dir):
            try:
                os.makedirs(master_dir, exist_ok=True)
            except Exception as e:
                env.logger.debug(
                    f"Failed to create ssh control master directory {master_dir}: {e}"
                )
                return ""
        return f"-o 'ControlMaster=auto' -o 'ControlPath={master_dir}/%r@%h:%p' -o 'ControlPersist=10m'"

    def _get_identify_file_options(self):
        if ("pem_file" in self.config and
                isinstance(self.config["pem_file"], str) and
                self.config["pem_file"].strip() != ""):
            return f""" -i '{self.config["pem_file"]}' """
        return ""

    def _get_send_cmd(self, rename=False):
        if rename:
            return (
                "ssh " + self.cm_opts + self.pem_opts +
                """ -q {host} -p {port} "mkdir -p {dest:dpq}" && """ +
                """rsync -a --no-g -e 'ssh """ + self.cm_opts + self.pem_opts +
                """ -p {port}' {source:aep} "{host}:{dest:dep}" && """ +
                """ssh """ + self.cm_opts + self.pem_opts +
                """ -q {host} -p {port} "mv {dest:dep}/{source:b} {dest:ep}" """
            )
        return ("ssh " + self.cm_opts + self.pem_opts +
                """ -q {host} -p {port} "mkdir -p {dest:dpq}" """ +
                """ && rsync -a --no-g -e "ssh -p {port} """ + self.cm_opts +
                self.pem_opts + ''' " {source:aep} "{host}:{dest:dep}"''')

    def _get_receive_cmd(self, rename=False):
        if rename:
            return ("""rsync -a --no-g -e 'ssh """ + self.cm_opts +
                    self.pem_opts +
                    """ -p {port}' {host}:{source:e} "{dest:adep}" && """ +
                    '''mv "{dest:adep}/{source:b}" "{dest:aep}"''')
        return ("""rsync -a --no-g -e 'ssh """ + self.cm_opts + self.pem_opts +
                ''' -p {port}' {host}:{source:e} "{dest:adep}"''')

    def _get_execute_cmd(self, under_workdir=True, use_heredoc=True) -> str:
        # #1396
        if "execute_cmd" in self.config:
            return self.config["execute_cmd"]
        if use_heredoc:
            return (
                "ssh " + self.cm_opts + self.pem_opts +
                """ -q {host} -p {port} <<'HEREDOC!!'\nbash --login -c '""" +
                (" [ -d {workdir} ] || mkdir -p {workdir}; cd {workdir} && "
                 if under_workdir else " ") + """ {cmd} '\nHEREDOC!!\n""")
        else:
            return ("ssh " + self.cm_opts + self.pem_opts +
                    """ -q {host} -p {port} "bash --login -c '""" +
                    (" [ -d {workdir} ] || mkdir -p {workdir}; cd {workdir} && "
                     if under_workdir else " ") + """ {cmd}'" """)

    def _get_query_cmd(self):
        return self.config.get(
            "query_cmd",
            """ssh """ + self.cm_opts + self.pem_opts +
            """ -q {host} -p {port} "bash --login -c '{self.config.get("sos", "sos")} status {task} -v 0'" """,
        )

    def is_shared(self, path):
        fullpath = os.path.abspath(os.path.expanduser(path))
        for sdir in self.shared_dirs:
            if fullpath.startswith(sdir):
                # issue 710, if a directory is both under path_map and shared, then it is not considered to be shared.
                if not any(
                        fullpath.startswith(mdir)
                        for mdir in self.path_map.keys()):
                    return True
        return False

    def _map_path(self, source):
        result = {}
        cwd = os.getcwd()
        if isinstance(source, str) and source.startswith("#"):
            result[path(source, host="localhost")] = path(
                source, host=self.alias)
        elif isinstance(source, (str, path)):
            dest = os.path.abspath(os.path.expanduser(source))
            # we use samefile to avoid problems with case-insensitive file system #522
            # we also use the "cwd" name to avoid wrong case for cwd. For example,
            # if the cwd = '/Users/user/Project'
            # then, dest = '/USERS/USER/PROJECT/a.txt'
            # would be converted to '/Users/user/Project/a.txt' before path mapping
            if os.path.exists(dest[:len(cwd)]) and os.path.samefile(
                    dest[:len(cwd)], cwd):
                dest = cwd + dest[len(cwd):]
            matched = [
                k for k in self.path_map.keys()
                if os.path.exists(dest[:len(k)]) and
                os.path.samefile(dest[:len(k)], k)
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = self.path_map[k] + dest[len(k):]
            else:
                env.logger.debug(
                    f"Path {source} is not under any specified paths of localhost and is mapped to {dest} on remote host."
                )
            result[source] = dest.replace("\\", "/")
        elif isinstance(source, (Sequence, sos_targets)):
            for src in source:
                result.update(self._map_path(src))
        else:
            env.logger.debug(f"Ignore unmappable source {source}")
            return {source: source}
        return result

    #
    # Interface functions
    #
    def _map_var(self, source):
        cwd = os.getcwd()
        if isinstance(source, path):
            source = str(source)
        if isinstance(source, str):
            if source.startswith("#"):
                return path(source, host=self.alias)
            dest = os.path.abspath(os.path.expanduser(source))
            # we use samefile to avoid problems with case-insensitive file system #522
            # we also use the "cwd" name to avoid wrong case for cwd. For example,
            # if the cwd = '/Users/user/Project'
            # then, dest = '/USERS/USER/PROJECT/a.txt'
            # would be converted to '/Users/user/Project/a.txt' before path mapping
            if os.path.exists(dest[:len(cwd)]) and os.path.samefile(
                    dest[:len(cwd)], cwd):
                dest = cwd + dest[len(cwd):]
            matched = [
                k for k in self.path_map.keys()
                if os.path.exists(dest[:len(k)]) and
                os.path.samefile(dest[:len(k)], k)
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = self.path_map[k] + dest[len(k):]
            else:
                env.logger.debug(
                    f"Path {source} is not under any specified paths of localhost and is mapped to {dest} on remote host."
                )
            return dest.replace("\\", "/")
        elif isinstance(source, (Sequence, set, sos_targets)):
            ret = [self._map_var(x) for x in source]
            return [x for x in ret if x is not None]
        else:
            env.logger.debug(f"Ignore unmappable source {source}")
            return source

    def test_connection(self):
        try:
            cmd = cfg_interpolate(
                "ssh {host} {pf_opt} -p {port} true",
                {
                    "host": self.address,
                    "port": self.port,
                    "pf_opt": f"-i '{self.pem_file}'" if self.pem_file else "",
                },
            )
            p = pexpect.spawn(cmd)
            # could be prompted for Password or password, so use assword
            while True:
                i = p.expect(
                    [
                        "(?i)are you sure you want to continue connecting",
                        "[pP]assword:",
                        pexpect.EOF,
                    ],
                    timeout=5,
                )
                if i == 0:
                    p.sendline("yes")
                elif i == 1:
                    p.close(force=True)
                    from .remote import stty_sane

                    stty_sane()
                    return f"ssh connection to {self.address} was prompted for password. Please set up public key authentication to the remote host before continue."
                elif i == 2:
                    p.close()
                    if p.exitstatus == 0:
                        return "OK"
                    elif p.before:
                        return p.before.decode()
                    else:
                        return f'Command "{cmd}" exits with code {p.exitstatus}'
        except pexpect.TIMEOUT:
            return f"ssh connection to {self.address} time out with prompt: {str(p.before)}"
        except Exception as e:
            return f"Failed to check remote connection {self.address}:{self.port}: {e}"
        return "OK"

    def _reverse_map_var(self, dest):
        if isinstance(dest, path):
            dest = str(dest)
        if isinstance(dest, str):
            matched = [
                l for l, r in self.path_map.items()
                if dest.startswith(r) and (len(r) == len(dest) or r.endswith(
                    "/") or r.endswith("\\") or dest[len(r)] in ("/", "\\"))
            ]
            if matched:
                # pick the longest key that matches
                k = max(matched, key=len)
                dest = k + dest[len(self.path_map[k]):]
            else:
                env.logger.debug(
                    f"Path {dest} is not under any specified paths of localhost and is mapped to {dest} on local host."
                )
            return dest.replace("\\", "/")
        elif isinstance(dest, (Sequence, set, sos_targets)):
            ret = [self._reverse_map_var(x) for x in dest]
            return [x for x in ret if x is not None]
        else:
            env.logger.debug(f"Ignore unmappable source {dest}")
            return dest

    def _remote_abs(self, path):
        # return an absolute path relative to remote host
        path = str(path)
        if os.path.isabs(path):
            return path
        return os.path.join(self._map_var(os.getcwd()), path)

    def send_to_host(self, items):
        # we only copy files and directories, not other types of targets
        if isinstance(items, str):
            p = [path(items)]
        elif isinstance(items, path):
            p = [items]
        elif isinstance(items, Sequence):
            ignored = [x for x in items if not isinstance(x, (str, path))]
            if ignored:
                env.logger.info(f"``Ignore`` {ignored}")
            p = [path(x) for x in items if isinstance(x, (str, path))]
        else:
            env.logger.warning(
                f"Unrecognized items to be sent to host: {items}")
            return {}

        items = sum(
            [list(x.parent.glob(x.name)) for x in p],
            [],
        )

        from .utils import find_symbolic_links

        new_items = []
        for item in items:
            links = find_symbolic_links(item)
            for link, realpath in links.items():
                env.logger.info(f"Adding {realpath} for symbolic link {link}")
            new_items.extend(links.values())
        items.extend(new_items)

        sending = self._map_path(items)

        sent = {}
        for source in sorted(sending.keys()):
            dest = self._remote_abs(sending[source])
            if self.is_shared(source):
                if ("TASK" in env.config["SOS_DEBUG"] or
                        "ALL" in env.config["SOS_DEBUG"]):
                    env.log_to_file(
                        "TASK", f"Skip sending {source} on shared file system")
            else:
                if ("TASK" in env.config["SOS_DEBUG"] or
                        "ALL" in env.config["SOS_DEBUG"]):
                    env.log_to_file(
                        "TASK", f"Sending ``{source}`` to {self.alias}:{dest}")
                cmd = cfg_interpolate(
                    self._get_send_cmd(
                        rename=os.path.basename(source) != os.path.basename(
                            dest)),
                    {
                        "source": sos_targets(str(source).rstrip("/")),
                        "dest": sos_targets(dest),
                        "host": self.address,
                        "port": self.port,
                    },
                )
                if ("TASK" in env.config["SOS_DEBUG"] or
                        "ALL" in env.config["SOS_DEBUG"]):
                    env.log_to_file("TASK", cmd)
                ret = subprocess.call(
                    cmd,
                    shell=True,
                    stderr=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                )
                if ret != 0:
                    raise RuntimeError(
                        f'Failed to copy {source} to {self.alias} using command "{cmd}". The remote host might be unavailable.'
                    )
            sent[source] = dest
        return sent

    def receive_from_host(self, items):
        if isinstance(items, str):
            items = [items]
        elif isinstance(items, path):
            items = [str(items)]
        elif isinstance(items, Sequence):
            ignored = [x for x in items if not isinstance(x, (str, path))]
            if ignored:
                env.logger.info(f"``Ignore`` {ignored}")
            items = [x for x in items if isinstance(x, (str, path))]
        else:
            env.logger.warning(
                f"Unrecognized items to be retrieved from host: {items}")
            return {}

        # y could be path
        receiving = {
            self._remote_abs(y): str(x)
            for x, y in self._map_path(items).items()
        }
        #
        received = {}
        for source in sorted(receiving.keys()):
            dest = receiving[source]
            dest_dir = os.path.dirname(dest)
            if dest_dir and not os.path.isdir(dest_dir):
                try:
                    os.makedirs(dest_dir)
                except Exception as e:
                    env.logger.error(
                        f"Failed to create destination directory {dest_dir}: {e}"
                    )
            if self.is_shared(dest) and os.path.basename(
                    source) == os.path.basename(dest):
                env.logger.debug(
                    f"Skip retrieving ``{dest}`` from shared file system")
                received[dest] = source
            else:
                cmd = cfg_interpolate(
                    self._get_receive_cmd(
                        rename=os.path.basename(source) != os.path.basename(
                            dest)),
                    {
                        "source": sos_targets(str(source).rstrip("/")),
                        "dest": sos_targets(dest),
                        "host": self.address,
                        "port": self.port,
                    },
                )
                if ("TASK" in env.config["SOS_DEBUG"] or
                        "ALL" in env.config["SOS_DEBUG"]):
                    env.log_to_file("TASK", cmd)
                try:
                    ret = subprocess.call(
                        cmd,
                        shell=True,
                        stderr=subprocess.DEVNULL,
                        stdout=subprocess.DEVNULL,
                    )
                    if ret != 0:
                        raise RuntimeError(f"command return {ret}")
                    received[dest] = source
                except Exception as e:
                    raise RuntimeError(
                        f'Failed to copy {source} from {self.alias} using command "{cmd}": {e}'
                    )
        return received

    #
    # Interface
    #
    def prepare_task(self, task_id):
        try:
            self._prepare_task(task_id)
            return True
        except Exception as e:
            env.logger.error(str(e))
            return False

    def _prepare_task(self, task_id):
        task_file = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".task")
        if not os.path.isfile(task_file):
            raise ValueError(f"Missing task definition {task_file}")
        tf = TaskFile(task_id)
        params, old_runtime = tf.get_params_and_runtime()
        task_vars = params.sos_dict
        runtime = {
            "_runtime": {
                x: task_vars["_runtime"][x] for x in (
                    "verbosity",
                    "sig_mode",
                    "run_mode",
                    "walltime",
                    "cores",
                    "mem",
                ) if x in task_vars["_runtime"]
            },
            task_id: {},
        }

        if (self.config.get("max_mem", None) is not None and
                task_vars["_runtime"].get("mem", None) is not None and
                self.config["max_mem"] < task_vars["_runtime"]["mem"]):
            raise ValueError(
                f'Task {task_id} requested more mem ({task_vars["_runtime"]["mem"]}) than allowed max_mem ({self.config["max_mem"]})'
            )
        if (self.config.get("max_cores", None) is not None and
                task_vars["_runtime"].get("cores", None) is not None and
                self.config["max_cores"] < task_vars["_runtime"]["cores"]):
            raise ValueError(
                f"Task {task_id} requested more cores ({task_vars['_runtime']['cores']}) than allowed max_cores ({self.config['max_cores']})"
            )
        if (self.config.get("max_walltime", None) is not None and
                task_vars["_runtime"].get("walltime", None) is not None and
                expand_time(self.config["max_walltime"]) < expand_time(
                    task_vars["_runtime"]["walltime"])):
            raise ValueError(
                f'Task {task_id} requested more walltime ({task_vars["_runtime"]["walltime"]}) than allowed max_walltime ({self.config["max_walltime"]})'
            )

        if task_vars["_input"] and not isinstance(task_vars["_input"],
                                                  Undetermined):
            sent = self.send_to_host(task_vars["_input"])
            if sent:
                env.logger.info(
                    f"{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}"
                )
        if task_vars["_depends"] and not isinstance(task_vars["_depends"],
                                                    Undetermined):
            sent = self.send_to_host(task_vars["_depends"])
            if sent:
                env.logger.info(
                    f"{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}"
                )
        if "to_host" in task_vars["_runtime"]:
            if not isinstance(
                    task_vars["_runtime"]["to_host"],
                (str, Sequence)) or (isinstance(
                    task_vars["_runtime"]["to_host"], Sequence) and not all(
                        isinstance(x, str)
                        for x in task_vars["_runtime"]["to_host"])):
                raise ValueError(
                    f'Parameter to_host accepts a list of paths (strings). {task_vars["_runtime"]["to_host"]} provided'
                )

            sent = self.send_to_host(task_vars["_runtime"]["to_host"])
            if sent:
                env.logger.info(
                    f"{task_id} ``sent`` {short_repr(sent.keys())} to {self.alias}"
                )

        # map variables
        runtime["_runtime"]["workdir"] = (
            task_vars["_runtime"]["workdir"] if "workdir"
            in task_vars["_runtime"] else path.cwd().to_named_path())

        if runtime["_runtime"]["workdir"].startswith("#"):
            try:
                path(runtime["_runtime"]["workdir"], host=self.alias)
            except Exception as e:
                raise ValueError(
                    f'Working directory {runtime["_runtime"]["workdir"]} does not exist on remote host {self.alias}: {e}'
                )
        elif path(runtime["_runtime"]["workdir"]).is_absolute():
            env.logger.debug(
                f'Absolute path {path(runtime["_runtime"]["workdir"])} used as workdir.'
            )

        env.log_to_file("TASK",
                        f'Set workdir to {runtime["_runtime"]["workdir"]}')

        # server restrictions #488
        for key in ("max_mem", "max_cores", "max_walltime"):
            if key in self.config:
                runtime["_runtime"][key] = (
                    format_HHMMSS(self.config[key])
                    if key == "max_walltime" else self.config[key])
        runtime["_runtime"]["localhost"] = get_config(
            ["hosts", self.alias], allowed_keys=["shared", "paths"])
        # only update task file if there are runtime information
        if len(runtime) > 1 or runtime["_runtime"] or runtime != old_runtime:
            tf.runtime = runtime

        tf.status = "pending"
        self.send_job_file(task_file)

    def send_job_file(self, job_file, dir="tasks"):
        send_cmd = cfg_interpolate(
            f"ssh {self.cm_opts + self.pem_opts}"
            f' -q {{address}} -p {{port}} "[ -d ~/.sos/{dir} ] || mkdir -p ~/.sos/{dir}" && '
            f' rsync -a --no-g -e "ssh {self.cm_opts + self.pem_opts}'
            f' -q -p {{port}}" {{job_file:ap}} {{address}}:.sos/{dir}/',
            {
                "job_file": sos_targets(job_file),
                "address": self.address,
                "port": self.port,
            },
        )

        # use scp for this simple case
        try:
            subprocess.check_call(send_cmd, shell=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Failed to copy job {job_file} to {self.alias} using command {send_cmd}: {e}"
            )

    def check_output(self,
                     cmd: object,
                     under_workdir=False,
                     **kwargs) -> object:
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(
                    under_workdir=under_workdir, use_heredoc="." in cmd),
                {
                    "host": self.address,
                    "port": self.port,
                    "cmd": cmd.replace("'", r"'\''"),
                    "workdir": self._map_var(os.getcwd()),
                },
            )
        except Exception as e:
            raise ValueError(
                f'Failed to run command {cmd}: {e} ({env.sos_dict["CONFIG"]})')
        if "TASK" in env.config["SOS_DEBUG"] or "ALL" in env.config["SOS_DEBUG"]:
            env.log_to_file("TASK", f"Executing command ``{cmd}``")
        try:
            return subprocess.check_output(cmd, shell=True, **kwargs).decode()
        except Exception as e:
            env.logger.debug(f"Check output of {cmd} failed: {e}")
            raise

    def check_call(self, cmd, under_workdir=False, **kwargs):
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(
                    under_workdir=under_workdir, use_heredoc="." in cmd),
                {
                    "host": self.address,
                    "port": self.port,
                    "cmd": cmd.replace("'", r"'\''"),
                    "workdir": self._map_var(os.getcwd()),
                },
            )
        except Exception as e:
            raise ValueError(f"Failed to run command {cmd}: {e}")
        if "TASK" in env.config["SOS_DEBUG"] or "ALL" in env.config["SOS_DEBUG"]:
            env.log_to_file("TASK", f"Executing command ``{cmd}``")
        try:
            return subprocess.check_call(cmd, shell=True, **kwargs)
        except Exception as e:
            env.logger.debug(f"Check output of {cmd} failed: {e}")
            raise

    def run_command(self, cmd, wait_for_task, realtime=False, **kwargs):
        if isinstance(cmd, list):
            cmd = subprocess.list2cmdline(cmd)
        try:
            cmd = cfg_interpolate(
                self._get_execute_cmd(
                    under_workdir=False, use_heredoc="." in cmd),
                {
                    "host": self.address,
                    "port": self.port,
                    "cmd": cmd.replace("'", r"'\''"),
                    "workdir": self._map_var(os.getcwd()),
                },
            )
        except Exception as e:
            raise ValueError(f"Failed to run command {cmd}: {e}")
        if "TASK" in env.config["SOS_DEBUG"] or "ALL" in env.config["SOS_DEBUG"]:
            env.log_to_file("TASK", f"Executing command ``{cmd}``")
        if realtime:
            from .utils import pexpect_run

            return pexpect_run(cmd)
        elif wait_for_task or sys.platform == "win32":
            # keep proc persistent to avoid a subprocess is still running warning.
            p = subprocess.Popen(cmd, shell=True, **kwargs)
            p.wait()
        else:
            p = DaemonizedProcess(cmd, **kwargs)
            p.start()
            p.join()

    def receive_result(self, task_id: str) -> Dict[str, int]:
        # for filetype in ('res', 'status', 'out', 'err'):
        sys_task_dir = os.path.join(os.path.expanduser("~"), ".sos", "tasks")
        # use -p to preserve modification times so that we can keep the job status locally.
        receive_cmd = cfg_interpolate(
            "scp -P {port} {pem_opts} -p -q {address}:.sos/tasks/{task}.* {sys_task_dir}",
            {
                "port": self.port,
                "address": self.address,
                "task": task_id,
                "sys_task_dir": sys_task_dir,
                "pem_opts": self.pem_opts,
            },
        )
        # it is possible that local files are readonly (e.g. a pluse file) so we first need to
        # make sure the files are readable and remove them. Also, we do not want any file that is
        # obsolete to appear as new after copying
        for lfile in glob.glob(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "tasks", task_id + ".*")):
            if not os.access(lfile, os.W_OK):
                os.chmod(lfile, stat.S_IREAD | stat.S_IWRITE)
            # os.remove(lfile)
        if "TASK" in env.config["SOS_DEBUG"] or "ALL" in env.config["SOS_DEBUG"]:
            env.log_to_file("TASK", receive_cmd)
        ret = subprocess.call(
            receive_cmd,
            shell=True,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
        if ret != 0:
            # this time try to get error message
            ret = subprocess.call(receive_cmd, shell=True)
            if ret != 0:
                raise RuntimeError(
                    "Failed to retrieve result of job {} from {} with cmd\n{}"
                    .format(task_id, self.alias, receive_cmd))

        tf = TaskFile(task_id)
        params = tf.params
        res = tf.result

        if not res:
            env.logger.debug(
                f"Result for {task_id} is not received (no result)")
            return {
                "ret_code": 1,
                "task": task_id,
                "exception": RuntimeError(f"Task {task_id} failed or aborted"),
                "output": sos_targets(),
            }

        if ("ret_code" in res and res["ret_code"] != 0) or ("succ" in res and
                                                            res["succ"] != 0):
            _show_err_and_out(task_id, res)
            env.logger.info(f"Ignore remote results for failed job {task_id}")
            return res

        if env.verbosity >= 3:
            _show_err_and_out(task_id, res)
        # do we need to copy files? We need to consult original task file
        # not the converted one
        job_dict = params.sos_dict
        if ("_output" in job_dict and job_dict["_output"] and
                not isinstance(job_dict["_output"], Undetermined) and
                env.config["run_mode"] != "dryrun"):
            received = self.receive_from_host(
                [x for x in job_dict["_output"] if isinstance(x, (str, path))])
            if received:
                env.logger.info(
                    f"{task_id} ``received`` {short_repr(received.keys())} from {self.alias}"
                )
        if "from_host" in job_dict[
                "_runtime"] and env.config["run_mode"] != "dryrun":
            if isinstance(job_dict["_runtime"]["from_host"], (Sequence, str)):
                received = self.receive_from_host(
                    job_dict["_runtime"]["from_host"])
                if received:
                    env.logger.info(
                        f"{task_id} ``received`` {short_repr(received.keys())} from {self.alias}"
                    )
            else:
                env.logger.warning(
                    f"Expecting a string or list of string from from_host: {job_dict['_runtime']['from_host']} received"
                )
        # we need to translate result from remote path to local
        if "output" in res:
            if "_output" not in job_dict:
                # this should exist, but let us check it for robustness
                env.logger.warning("Missing _output in task dict")
                res["output"] = sos_targets()
            elif job_dict["_output"].undetermined():
                res["output"] = sos_targets(
                    self._reverse_map_var(res["output"]))
            else:
                res["output"] = job_dict["_output"]
        if "subtasks" in res:
            for tid, subparams in params.task_stack:
                if tid in res["subtasks"] and "output" in res["subtasks"][tid]:
                    if "_output" not in subparams.sos_dict:
                        # this should not happen
                        env.logger.warning("Missing _output in subparams")
                        res["subtasks"][tid]["output"] = sos_targets()
                    elif subparams.sos_dict["_output"].undetermined():
                        res["subtasks"][tid]["output"] = sos_targets(
                            self._reverse_map_var(
                                res["subtasks"][tid]["output"]))
                    else:
                        res["subtasks"][tid]["output"] = subparams.sos_dict[
                            "_output"]
        return res


#
# host instances are shared by all tasks so there should be only one
# instance for each host.
#


class Host:
    host_instances: Dict = {}

    def __init__(
        self,
        alias: Optional[str] = "",
        start_engine: bool = True,
        test_connection: bool = True,
    ) -> None:
        # a host started from Jupyter notebook might not have proper stdout
        # (a StringIO) and cannot be used for DaemonizedFork).
        self._get_config(alias)
        self._get_host_agent(start_engine, test_connection)

    # for test purpose
    @classmethod
    def reset(cls) -> None:
        for host in cls.host_instances.values():
            # perhaps the engine failed to start
            if hasattr(host, "_task_engine"):
                del host._task_engine
        cls.host_instances = {}

    def _detect_host(self) -> str:
        hostname = socket.gethostname()
        ips = socket.gethostbyname_ex(hostname)[2]
        ips = [ip for ip in ips if not ip.startswith("127.")]
        hostname = hostname.lower()
        #
        for host, host_info in env.sos_dict["CONFIG"]["hosts"].items():
            # find by key hostname
            if "hostname" in host_info and host_info["hostname"].lower(
            ) == hostname:
                return host
            # find by key hostname
            if "hostname" in host_info:
                hn = get_config(
                    "hosts", host, "hostname", expected_type=str).lower()
                if hn.split(".")[0] == hostname or hn == hostname.split(".")[0]:
                    return host
            # find by alias
            if host.lower() == hostname:
                return host
            # find by address
            if "address" in host_info:
                addr = get_config("hosts", host, "address", expected_type=str)
                if (addr.split("@")[-1].lower() == hostname or addr.split(
                        ".", 1)[0].split("@")[-1].lower() == hostname):
                    return host
                if any(ip == addr.split("@")[-1] for ip in ips):
                    return host
        return hostname.split(".")[0]

    def _get_local_host(self) -> str:
        if "CONFIG" not in env.sos_dict or "hosts" not in env.sos_dict["CONFIG"]:
            from .utils import load_config_files
            load_config_files()
        # look for an entry with gethost
        if "hosts" not in env.sos_dict["CONFIG"]:
            env.sos_dict["CONFIG"]["hosts"] = {
                "localhost": {
                    "address": "localhost",
                    "alias": "localhost"
                }
            }
            return "localhost"
        #
        # check if a key localhost is defined
        if "localhost" in env.sos_dict["CONFIG"]:
            if (env.sos_dict["CONFIG"]["localhost"]
                    not in env.sos_dict["CONFIG"]["hosts"]):
                raise ValueError(
                    f"Undefined localhost {env.sos_dict['CONFIG']['localhost']}"
                )
            return env.sos_dict["CONFIG"]["localhost"]
        else:
            env.sos_dict["CONFIG"]["localhost"] = "localhost"
            return "localhost"

    def _get_remote_host(self, alias: Optional[str]) -> str:
        # get a remote host specified by Alias
        if "CONFIG" not in env.sos_dict or "hosts" not in env.sos_dict["CONFIG"]:
            from .utils import load_config_files

            load_config_files()
        if not alias or alias == "localhost":
            return self._get_local_host()
        if not isinstance(alias, str):
            raise ValueError(f"A string is expected for host {alias}")
        if "hosts" not in env.sos_dict["CONFIG"]:
            env.sos_dict["CONFIG"]["hosts"] = {}
        if alias in env.sos_dict["CONFIG"]["hosts"]:
            return alias
        # assuming the host name is a name or IP address
        env.logger.debug(
            f"Assuming {alias} to be a hostname or IP address not defined in hosts file"
        )
        env.sos_dict["CONFIG"]["hosts"][alias] = {
            "address": alias,
            "alias": alias,
        }
        return alias

    def _get_config(self, alias: Optional[str]) -> None:
        LOCAL = self._get_local_host()
        DETECTED = self._detect_host()
        REMOTE = self._get_remote_host(alias)
        self.alias = REMOTE

        if LOCAL == "localhost" and DETECTED in env.sos_dict["CONFIG"]["hosts"]:
            LOCAL = DETECTED

        # now we need to find definition for local and remote host
        if LOCAL == "localhost" and REMOTE == "localhost":
            self.config = {
                "address": "localhost",
                "alias": "localhost",
            }
        elif "hosts" in env.sos_dict["CONFIG"]:
            if LOCAL not in env.sos_dict["CONFIG"]["hosts"]:
                raise ValueError(f"Undefined local host {LOCAL}")
            if REMOTE not in env.sos_dict["CONFIG"]["hosts"]:
                raise ValueError(f"Undefined remote host {REMOTE}")

            # now we have definition for local and remote hosts but we only
            # exapnd address, which we have to know right now
            self.config = get_config(
                "hosts",
                self.alias,
                excluded_keys=("paths", "shared"),
                expand_keys=("address", "port", "pem_file"),
                expected_type=dict,
            )

            same_host = LOCAL == REMOTE
            if same_host and LOCAL != "localhost" and LOCAL != DETECTED and DETECTED not in env.sos_dict[
                    'CONFIG']['hosts']:
                # if "localhost" is defined, but does not match by ip address etc,
                # we assume that the matched_host is a separate host with the same
                # configuration (see #1407 for details)
                env.logger.debug(
                    f'Specified host {LOCAL} does not match detected host {DETECTED}.'
                )
                local_cfg = copy.deepcopy(
                    env.sos_dict["CONFIG"]["hosts"][LOCAL])
                env.sos_dict["CONFIG"]["hosts"][DETECTED] = local_cfg
                LOCAL = DETECTED

            cfg = env.sos_dict["CONFIG"]["hosts"]
            # if local and remote hosts are the same
            if (same_host or
                    "address" not in env.sos_dict["CONFIG"]["hosts"][REMOTE] or
                ("address" in env.sos_dict["CONFIG"]["hosts"][REMOTE] and
                 env.sos_dict["CONFIG"]["hosts"][REMOTE]["address"]
                 == "localhost")):
                # there would be no path map
                self.config["path_map"] = []
                self.config["shared"] = ["/"]
                # do not override address setting to use localhost
                # because the address should be used in #1407
                # self.config["address"] = "localhost"
            else:
                self.config["path_map"] = []

                def normalize_value(x):
                    x = cfg_interpolate(x)
                    return x if x.endswith(os.sep) else (x + os.sep)

                if "shared" in cfg[LOCAL] and "shared" in cfg[REMOTE]:
                    common = set(cfg[LOCAL]["shared"].keys()) & set(
                        cfg[REMOTE]["shared"].keys())
                    if common:
                        lcl_shrd = get_config(
                            "hosts", LOCAL, "shared", expected_type=dict)
                        rmt_shrd = get_config(
                            "hosts", REMOTE, "shared", expected_type=dict)
                        self.config["shared"] = [
                            normalize_value(lcl_shrd[x]) for x in common
                        ]
                        self.config["path_map"] = [
                            f"{normalize_value(lcl_shrd[x])} -> {normalize_value(rmt_shrd[x])}"
                            for x in common
                        ]
                # if paths are defined for both local and remote host, define path_map
                if ("paths" in cfg[LOCAL] and
                        cfg[LOCAL]["paths"]) and ("paths" in cfg[REMOTE] and
                                                  cfg[REMOTE]["paths"]):
                    if any(k not in cfg[REMOTE]["paths"]
                           for k in cfg[LOCAL]["paths"].keys()):
                        env.logger.debug(
                            f'One or more local paths {", ".join(cfg[LOCAL]["paths"].keys())} cannot be mapped to remote host {REMOTE} with paths {",".join(cfg[REMOTE]["paths"].keys())}'
                        )
                    #
                    lpaths = get_config(
                        "hosts", LOCAL, "paths", expected_type=dict)
                    rpaths = get_config(
                        "hosts", REMOTE, "paths", expected_type=dict)
                    self.config["path_map"].extend([
                        f"{normalize_value(lpaths[x])} -> {normalize_value(rpaths[x])}"
                        for x in lpaths.keys()
                        if x in rpaths
                    ])
                if "pem_file" in cfg[LOCAL]:
                    if isinstance(cfg[LOCAL]["pem_file"], dict):
                        if REMOTE in cfg[LOCAL]["pem_file"]:
                            self.config["pem_file"] = get_config(
                                "hosts",
                                LOCAL,
                                "pem_file",
                                REMOTE,
                                expected_type=str)
                    elif isinstance(cfg[LOCAL]["pem_file"], str):
                        self.config["pem_file"] = get_config(
                            "hosts", LOCAL, "pem_file", expected_type=str)
                    else:
                        raise ValueError(
                            f"Option pem_file should be a string or dictionary, {cfg[LOCAL]['pem_file']} provided."
                        )
        elif LOCAL == REMOTE:
            # now we have checked local and remote are not defined, but they are the same, so
            # it is safe to assume that they are both local hosts
            self.config = {
                "address": "localhost",
                "alias": LOCAL,
            }
        else:
            raise ValueError(
                f"Undefined local and remote hosts {LOCAL} and {REMOTE}.")
        #
        self.config["alias"] = self.alias
        self.description = self.config.get("description", "")

        # standardize parameters max_walltime, max_cores, and max_mem for the host
        if "max_walltime" in self.config:
            self.config["max_walltime"] = format_HHMMSS(
                self.config["max_walltime"])
        if "max_cores" in self.config:
            if not isinstance(self.config["max_cores"], int):
                raise ValueError("An integer is expected for max_cores")
        if "max_mem" in self.config:
            self.config["max_mem"] = expand_size(self.config["max_mem"])

    def _get_task_and_workflow_engine(self):
        if self._engine_type == "process":
            task_engine = BackgroundProcess_TaskEngine(
                self.host_instances[self.alias])
            workflow_engine = BackgroundProcess_WorkflowEngine(
                self.host_instances[self.alias])
        else:
            task_engine = None
            workflow_engine = None

            for entrypoint in pkg_resources.iter_entry_points(
                    group="sos_taskengines"):
                try:
                    if entrypoint.name == self._engine_type:
                        task_engine = entrypoint.load()(
                            self.host_instances[self.alias])
                        break
                except Exception as e:
                    env.logger.debug(
                        f"Failed to load task engine {self._engine_type}: {e}")

            for entrypoint in pkg_resources.iter_entry_points(
                    group="sos_workflowengines"):
                try:
                    if entrypoint.name == self._engine_type:
                        workflow_engine = entrypoint.load()(
                            self.host_instances[self.alias])
                        break
                except Exception as e:
                    env.logger.debug(
                        f"Failed to load workflow engine {self._engine_type}: {e}"
                    )

            if task_engine is None and workflow_engine is None:
                raise ValueError(
                    f'Failed to load task engine of type "{self._engine_type}". Please check the engine name or install relevant module.'
                )
        return task_engine, workflow_engine

    def _get_host_agent(self, start_engine: bool,
                        test_connection: bool) -> None:
        if "queue_type" not in self.config:
            self._engine_type = "process"
        else:
            self._engine_type = self.config["queue_type"].strip()
        # if there is no engine, or if the engine was stopped
        if self.alias not in self.host_instances or (
                hasattr(self.host_instances[self.alias], "_task_engine") and
                self.host_instances[self.alias]._task_engine._is_stopped):
            if self.config["address"] == "localhost":
                self.host_instances[self.alias] = LocalHost(
                    self.config, test_connection=test_connection)
            else:
                self.host_instances[self.alias] = RemoteHost(
                    self.config, test_connection=test_connection)

            task_engine, workflow_engine = self._get_task_and_workflow_engine()
            self.host_instances[self.alias]._task_engine = task_engine
            self.host_instances[self.alias]._workflow_engine = workflow_engine

            # # the task engine is a thread and will run continously
            # if start_engine and task_engine is not None:
            #     task_engine.start()

        self._host_agent = self.host_instances[self.alias]
        # for convenience
        if hasattr(self._host_agent, "_task_engine"):
            self._task_engine = self._host_agent._task_engine
        if hasattr(self._host_agent, "_workflow_engine"):
            self._workflow_engine = self._host_agent._workflow_engine
        # it is possible that Host() is initialized before with start_engine=False
        # and called again to start engine
        if (start_engine and self._task_engine is not None and
                not self._task_engine.is_alive()):
            self._task_engine.start()

    # public interface
    #
    # based on Host definition
    #
    def send_to_host(self, items):
        return self._host_agent.send_to_host(items)

    def receive_from_host(self, items):
        return self._host_agent.receive_from_host(items)

    def map_var(self, rvars):
        return self._host_agent._map_var(rvars)

    def target_exists(self, targets):
        return self._host_agent.target_exists(targets)

    def target_signature(self, targets):
        return self._host_agent.target_signature(targets)

    def submit_task(self, task_id: str) -> str:
        if not self._task_engine:
            raise RuntimeError(
                f"No task engine or invalid engine definition defined for host {self.alias}"
            )
        return self._task_engine.submit_task(task_id)

    def check_status(self, tasks: List[str]) -> List[str]:
        # find the task engine
        return [self._task_engine.check_task_status(task) for task in tasks]

    def retrieve_results(self, tasks: List[str]):
        return self._task_engine.get_results(tasks)

    def execute_workflow(self, script, cmd, **template_args):
        if not self._workflow_engine:
            raise RuntimeError(
                f"No workflow engine or invalid engine definition defined for host {self.alias}"
            )
        return self._workflow_engine.execute_workflow(script, cmd,
                                                      **template_args)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import ast
import contextlib
import copy
import pickle
import sys
from typing import Any, Dict, Optional, Set

from ._version import __version__
from .utils import (ArgumentError, as_fstring, env, load_config_files,
                    pickleable)


def interpolate(text, global_dict=None, local_dict=None):
    """Evaluate expressions in `text` """
    # step 1, make it a f-string (add quotation marks and f
    # step 2, evaluate as a string
    try:
        return eval(as_fstring(text), global_dict, local_dict)
    except Exception as e:
        raise ValueError(f"Failed to interpolate {text}: {e}")


def cfg_interpolate(text, local_dict={}):
    # handle nested interpolate ...
    # we need to avoid containing our CONFIG with all the modules created by eval
    cfg_dict = copy.deepcopy(env.sos_dict.get("CONFIG", {}))
    if "os.environ" in text:
        exec("import os", cfg_dict)
    while True:
        res = interpolate(text, cfg_dict, local_dict)
        if res == text:
            break
        else:
            text = res
    return res


def get_config(*args, **kwargs):
    """
    Obtain configuration from sos configuration files, by default with variable intepolation.

    The basic way to use this function is to do

       get_config('key')

    which gets the value associated with configuration "key". If the value is a dictionary, you
    can go into the dictionary with any of the following

       get_config('key', 'key1')
       get_config(['key', 'key1'])
       get_config('key.key1')

    The expressions are by default interpolated with variables defined in the order of,

    1. specified from get_config
    2. the same dictionary in which the expression is evaluated
    3. the root of the configuration dictionary

    You can specify variables as a dictionary

        get_config('a', {'val': 5})

    if the key does not conflict with other keywords, you can specify directly as keyword arguments

        get_config('a', val=5)

    The funciton accepts the following keyword parameters:

      default: default value to return if the key is not found. Default to None.
      raw: return raw values without interpolation.
      raw_keys: keys that are not expanded
      expand_keys: keys that are expanded
      allowed_keys: retrieve only specified keys from a dicitonary.
      exclude_keys: exclude specified keys
      expected_type: expected type or types
    """
    default = kwargs.get("default", None)
    allowed_keys = kwargs.get("allowed_keys", None)
    excluded_keys = kwargs.get("excluded_keys", None)
    raw_keys = kwargs.get("raw_keys", None)
    expand_keys = kwargs.get("expand_keys", None)
    raw = kwargs.get("raw", False)
    expected_type = kwargs.get("expected_type", None)

    #
    keys = []
    custom_dict = {}
    for arg in args:
        if isinstance(arg, str):
            keys.append(arg)
        elif isinstance(arg, (list, tuple)):
            keys.extend(list(arg))
        elif isinstance(arg, dict):
            custom_dict.update(arg)
        else:
            raise ValueError(f"Unacceptable parameter {arg} for get_config.")
    #
    custom_dict.update({
        x: y for x, y in kwargs.items() if x not in (
            "default",
            "allowed_keys",
            "excluded_keys",
            "raw_keys",
            "expand_keys",
            "expected_type",
            "raw",
        )
    })
    #
    local_dict = {}
    val = env.sos_dict.get("CONFIG", {})
    while True:
        if not keys:
            break
        if not isinstance(val, dict):
            raise ValueError(
                f'A dictionary is expected to get item with key {".".join(keys)}: {val} obtained.'
            )
        local_dict = val
        key = keys[0]
        if key in val:
            keys.pop(0)
            val = val[key]
            continue
        elif "." in key:
            found = False
            subkeys = key.split(".")
            for j in range(len(subkeys)):
                subkey = ".".join(subkeys[:j + 1])
                if subkey in val:
                    found = True
                    val = val[subkey]
                    keys[0] = ".".join(subkeys[j + 1:])
                    if not keys[0]:
                        keys.pop(0)
                    break
            if not found:
                val = default
        else:
            val = default

    if expected_type is not None and not isinstance(val, expected_type):
        raise ValueError(
            f'A value of type {expected_type} is expected, "{val}" specified.')

    if raw is True:
        return val

    if isinstance(val, str):
        local_dict.update(custom_dict)
        return cfg_interpolate(val, local_dict)
    elif isinstance(val, dict):

        def interpolate_dict(
            item,
            custom_dict,
            allowed_keys=None,
            excluded_keys=None,
            raw_keys=None,
            expand_keys=None,
        ):
            res = {}
            for k, v in item.items():
                if allowed_keys and k not in allowed_keys:
                    continue
                if excluded_keys and k in excluded_keys:
                    continue
                if raw_keys and k in raw_keys:
                    res[k] = v
                    continue
                if expand_keys and k not in expand_keys:
                    res[k] = v
                    continue
                if isinstance(v, dict):
                    # v should be processed in place
                    res[k] = interpolate_dict(v, custom_dict)
                elif isinstance(v, str):
                    if "{" in v and "}" in v:
                        local_dict = copy.deepcopy(item)
                        local_dict.update(custom_dict)
                        res[k] = cfg_interpolate(v, local_dict)
                    else:
                        res[k] = v
                else:
                    res[k] = v
            return res

        return interpolate_dict(
            val,
            custom_dict,
            allowed_keys=allowed_keys,
            excluded_keys=excluded_keys,
            raw_keys=raw_keys,
            expand_keys=expand_keys,
        )
    else:
        return val


def get_accessed(node):
    """Get names, but ignore variables names to the left hand side
    That is to say, in case of
       a = b + 1
    we consider b as "being accessed", while a is not.
    """
    if isinstance(node, ast.Assign):
        return get_accessed(node.value)
    elif isinstance(node, ast.Name):
        return {node.id}
    names = set()
    if isinstance(node, list):
        for x in node:
            names |= get_accessed(x)
    else:
        for x in ast.iter_child_nodes(node):
            names |= get_accessed(x)
    return names


def accessed_vars(statement: str, mode: str = "exec") -> Set[str]:
    """Parse a Python statement and analyze the symbols used. The result
    will be used to determine what variables a step depends upon."""
    try:
        if mode == "exec":
            return get_accessed(ast.parse(statement, "<string>", "exec"))
        else:
            res = get_accessed(
                ast.parse("__NULL__(" + statement + ")", "<string>", "eval"))
            res.remove("__NULL__")
            return res
    except Exception:
        raise RuntimeError(
            f"Failed to parse statement: {statement} in {mode} mode")


def get_used_in_func(node):
    """Get names, but ignore variables names to the left hand side
    That is to say, in case of
       a = b + 1
    we consider b as "being accessed", while a is not.
    """
    if isinstance(node, ast.FunctionDef):
        return {node.name: get_accessed(node.body)}
    names = {}
    for node in ast.iter_child_nodes(node):
        names.update(get_used_in_func(node))
    return names


def used_in_func(statement: str,
                 filename: str = "<string>",
                 mode: str = "exec"):
    """Parse a Python statement and analyze the symbols used. The result
    will be used to determine what variables a step depends upon."""
    try:
        return get_used_in_func(ast.parse(statement, filename, mode))
    except Exception as e:
        raise RuntimeError(f"Failed to parse statement: {statement} {e}")


def SoS_eval(expr: str, extra_dict: dict = {}) -> Any:
    """Evaluate an expression with sos dict."""
    return eval(expr, env.sos_dict.dict(), extra_dict)


def _is_expr(expr):
    try:
        compile(expr, "<string>", "eval")
        return True
    except Exception:
        return False


class StatementHash(object):
    stmt_hash = {}

    def __init__(self) -> None:
        pass

    def hash(self, script: str) -> str:
        h = hash(script) & sys.maxsize
        StatementHash.stmt_hash[h] = script
        return f"script_{h}"

    def script(self, hash: str) -> str:
        return StatementHash.stmt_hash[int(hash[7:])]


stmtHash = StatementHash()


def SoS_exec(script: str,
             _dict: dict = None,
             return_result: bool = True) -> None:
    """Execute a statement."""
    if _dict is None:
        _dict = env.sos_dict.dict()

    if not return_result:
        if env.verbosity == 0:
            with contextlib.redirect_stdout(None):
                exec(
                    compile(
                        script, filename=stmtHash.hash(script), mode="exec"),
                    _dict)
        else:
            exec(
                compile(script, filename=stmtHash.hash(script), mode="exec"),
                _dict)
        return None

    try:
        stmts = list(ast.iter_child_nodes(ast.parse(script)))
        if not stmts:
            return
        if isinstance(stmts[-1], ast.Expr):
            # the last one is an expression and we will try to return the results
            # so we first execute the previous statements
            if len(stmts) > 1:
                if env.verbosity == 0:
                    with contextlib.redirect_stdout(None):
                        exec(
                            compile(
                                ast.Module(body=stmts[:-1], type_ignores=[]),
                                filename=stmtHash.hash(script),
                                mode="exec",
                            ),
                            _dict,
                        )
                else:
                    exec(
                        compile(
                            ast.Module(body=stmts[:-1], type_ignores=[]),
                            filename=stmtHash.hash(script),
                            mode="exec",
                        ),
                        _dict,
                    )
            # then we eval the last one
            if env.verbosity == 0:
                with contextlib.redirect_stdout(None):
                    res = eval(
                        compile(
                            ast.Expression(body=stmts[-1].value),
                            filename=stmtHash.hash(script),
                            mode="eval",
                        ),
                        _dict,
                    )
            else:
                res = eval(
                    compile(
                        ast.Expression(body=stmts[-1].value),
                        filename=stmtHash.hash(script),
                        mode="eval",
                    ),
                    _dict,
                )
        else:
            # otherwise we just execute the entire code
            if env.verbosity == 0:
                with contextlib.redirect_stdout(None):
                    exec(
                        compile(
                            script, filename=stmtHash.hash(script),
                            mode="exec"),
                        _dict,
                    )
            else:
                exec(
                    compile(
                        script, filename=stmtHash.hash(script), mode="exec"),
                    _dict)
            res = None
    except SyntaxError as e:
        raise SyntaxError(f"Invalid code {script}: {e}")

    # if check_readonly:
    #    env.sos_dict.check_readonly_vars()
    return res


#
# dynamic expression that cannot be resolved during parsing
# at prepare mode etc, and has to be resolved at run time.
#


class Undetermined(object):

    def __init__(self, expr: str = "") -> None:
        if not isinstance(expr, str):
            raise RuntimeError(
                f'Undetermined expression has to be a string: "{expr}" passed')
        self.expr = expr.strip()

    def value(self):
        return SoS_eval(self.expr)

    def __repr__(self) -> str:
        return f"Undetermined({self.expr!r})"

    def __hash__(self):
        raise RuntimeError(
            "Undetermined expression should be evaluated before used. "
            "This is certainly a bug so please report this to SoS developer.")

    def targets(self) -> "Undetermined":
        return self


class on_demand_options(object):
    """Expression that will be evaluated upon request."""

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        self._expressions = {}
        if items:
            self._expressions.update(items)

    def set(self, key: str, value: Any) -> None:
        self._expressions[key] = repr(value)

    def __contains__(self, key: str) -> bool:
        return key in self._expressions

    def __setitem__(self, key: str, value: str) -> None:
        self._expressions[key] = value

    def __getitem__(self, key: str) -> Any:
        # first check if the value if cached
        if key not in self._expressions:
            raise KeyError(key)
        try:
            return SoS_eval(self._expressions[key])
        except Exception as e:
            if key == "skip":
                raise ValueError(
                    f"Failed to evaluate option {key} with value {self._expressions[key]}: Only constant values are allowed for section option skip"
                )
            else:
                raise ValueError(
                    f"Failed to evaluate option {key} with value {self._expressions[key]}: {e}"
                )

    def __repr__(self):
        return repr(self._expressions)


class KeepOnlyImportAndDefine(ast.NodeTransformer):

    def __init__(self):
        self.level = 0

    def generic_visit(self, node):
        self.level += 1
        if self.level == 2 and not isinstance(
                node,
            (ast.Import, ast.ImportFrom, ast.FunctionDef, ast.ClassDef)):
            # print(f'remove {node}')
            ret = None
        else:
            ret = super(KeepOnlyImportAndDefine, self).generic_visit(node)
        self.level -= 1
        return ret


def analyze_global_statements(global_stmt):
    # find all import and function definition ...
    env.sos_dict.clear()
    env.sos_dict.set("SOS_VERSION", __version__)
    env.sos_dict.set("master_id", "")
    env.sos_dict.set("workflow_id", "")
    env.sos_dict.set("step_name", "")
    env.sos_dict.set("step_id", "")
    # first load CONFIG, this will create CONFIG
    load_config_files()

    # run only import, def, and class of the global_def
    transformer = KeepOnlyImportAndDefine()
    global_def = transformer.visit(
        ast.parse("from sos.runtime import *\n" + global_stmt))
    exec(compile(global_def, filename="<ast>", mode="exec"), env.sos_dict._dict)
    defined_keys = set(env.sos_dict.keys())

    # execute the entire statement
    try:
        SoS_exec(global_stmt)
    except ArgumentError:
        raise
    except Exception as e:
        raise RuntimeError(f"Failed to execute global statement: {e}")
    #
    global_vars = {
        k: env.sos_dict[k] for k in (set(env.sos_dict.keys()) - defined_keys)
        | {"SOS_VERSION", "CONFIG"}
    }
    # test if global vars can be pickled
    try:
        pickle.dumps(global_vars)
    except Exception:
        for key in set(env.sos_dict.keys()) - defined_keys:
            if not pickleable(env.sos_dict[key], key):
                raise ValueError(
                    f"Variable {key} cannot be defined in global section because it cannot be pickled to workers."
                )
    return global_def, global_vars



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

#
# Utility functions used by various executors.
#
import os
import re
import sys
import traceback
from collections.abc import Sequence
from io import StringIO
from tokenize import generate_tokens
from typing import Any

import psutil

from .controller import request_answer_from_controller
from .eval import SoS_eval, analyze_global_statements, stmtHash
from .syntax import SOS_TAG
from .targets import (RemovedTarget, RuntimeInfo, dynamic, file_target,
                      sos_step, sos_targets, sos_variable)
from .tasks import TaskParams
from .utils import (Error, env, expand_size, format_HHMMSS, get_traceback,
                    load_config_files, textMD5)


class ExecuteError(Error):
    """An exception to collect exceptions raised during run time so that
    other branches of the DAG would continue if some nodes fail to execute."""

    def __init__(self, workflow: str) -> None:
        Error.__init__(self)
        self.workflow = workflow
        self.errors = []
        self.traces = []
        self.args = (workflow,)

    def append(self, line: str, error: Exception) -> None:
        lines = [x for x in line.split("\n") if x.strip()]
        if not lines:
            short_line = ""
        else:
            short_line = ("[" +
                          (lines[0][:40] if len(lines[0]) > 40 else lines[0]) +
                          "]:")
        self.errors.append(short_line)
        self.traces.append(get_traceback())
        newline = "\n" if self.message else ""
        self.message += f"{newline}{short_line} {error}"


def __null_func__(*args, **kwargs) -> Any:
    """This function will be passed to SoS's namespace and be executed
    to evaluate functions of input, output, and depends directives."""

    def _flatten(x):
        if isinstance(x, str):
            return [x]
        elif isinstance(x, sos_targets):
            return [x]
        elif isinstance(x, Sequence):
            return sum((_flatten(k) for k in x), [])
        elif hasattr(x, "__flattenable__"):
            return _flatten(x.flatten())
        else:
            return [x]

    return _flatten(args), kwargs


def __output_from__(
    steps,
    group_by="auto",
    paired_with=None,
    pattern=None,
    group_with=None,
    for_each=None,
    remove_empty_groups=False,
):
    targets = sos_targets()
    if isinstance(steps, (int, str)):
        steps = [steps]
    elif isinstance(steps, Sequence):
        steps = list(steps)
    else:
        raise ValueError(
            f"Unacceptable value of input prameter from: {steps} provided")
    #
    for step in steps:
        if isinstance(step, int):
            if step == -1:
                # this refers to the last step of a forward style workflow
                if "__last_step__" in env.sos_dict and env.sos_dict[
                        "__last_step__"]:
                    step = env.sos_dict["__last_step__"]
                else:
                    raise ValueError(
                        "output_from(-1) is called for a step without previous step"
                    )
            elif "_" in env.sos_dict["step_name"]:
                step = f"{env.sos_dict['step_name'].rsplit('_', 1)[0]}_{step}"
            else:
                step = str(step)
        res = request_answer_from_controller(["step_output", step])
        if res is None or not isinstance(res, sos_targets):
            raise RuntimeError(f"Failed to obtain output of step {step}")
        targets.extend(res)

    if group_by != "auto" or paired_with or pattern or group_with or for_each:
        targets = sos_targets(
            targets,
            group_by=group_by,
            paired_with=paired_with,
            pattern=pattern,
            group_with=group_with,
            for_each=for_each,
        )
    return targets._remove_empty_groups() if remove_empty_groups else targets


def __traced__(*args, **kwargs):
    return sos_targets(*args, **kwargs).set_traced()


def __named_output__(
    name,
    group_by="auto",
    paired_with=None,
    pattern=None,
    group_with=None,
    for_each=None,
    remove_empty_groups=False,
):
    targets = request_answer_from_controller(["named_output", name])
    if targets is None:
        env.logger.warning(f'named_output("{name}") is not found')
        return sos_targets([])

    if group_by != "auto" or paired_with or pattern or group_with or for_each:
        targets = sos_targets(
            targets,
            group_by=group_by,
            paired_with=paired_with,
            pattern=pattern,
            group_with=group_with,
            for_each=for_each,
        )
    return targets._remove_empty_groups() if remove_empty_groups else targets


def clear_output(output=None):
    """
    Remove file targets in `_output` when a step fails to complete
    """
    for target in env.sos_dict["_output"] if output is None else output:
        if isinstance(target, file_target) and target.exists():
            try:
                target.unlink()
            except Exception as e:
                env.logger.warning(f"Failed to remove {target}: {e}")


def get_traceback_msg(e):
    error_class = e.__class__.__name__
    tb = sys.exc_info()[-1]
    msg = ""
    for st in reversed(traceback.extract_tb(tb)):
        if st.filename.startswith("script_"):
            code = stmtHash.script(st.filename)
            line_number = st.lineno
            code = "\n".join([
                f'{"---->" if i+1 == line_number else "     "} {x.rstrip()}'
                for i, x in enumerate(code.splitlines())
            ][max(line_number - 3, 0):line_number + 3])
            msg += f"""\
{st.filename} in {st.name}
{code}
"""
    detail = e.args[0] if e.args else ""
    if msg:
        return f"""
---------------------------------------------------------------------------
{error_class:42}Traceback (most recent call last)
{msg}
{error_class}: {detail}"""
    else:
        return f"{error_class}: {detail}"


def prepare_env(gdef="", gvars={}, extra_vars={}, host="localhost"):
    """clear current sos_dict, execute global_def (definitions and imports),
    and inject global variables"""
    env.sos_dict.clear()

    if not gdef and not gvars:
        # SoS Notebook calls prepare_env without global statement from a
        # particular
        gdef, gvars = analyze_global_statements("")

    if gdef:
        exec(compile(gdef, filename="<ast>", mode="exec"), env.sos_dict._dict)

    env.sos_dict.quick_update(gvars)
    env.sos_dict.quick_update(extra_vars)
    if "CONFIG" not in env.sos_dict:
        # if this is in sos notebook
        load_config_files()
    if ("hosts" not in env.sos_dict["CONFIG"] and
            "localhost" not in env.sos_dict["CONFIG"]):
        env.sos_dict["CONFIG"]["localhost"] = "localhost"
        env.sos_dict["CONFIG"]["hosts"] = {
            "localhost": {
                "paths": {},
                "address": "localhost"
            }
        }
    # expose `paths` of localhost
    if host == "localhost":
        if "localhost" in env.sos_dict["CONFIG"]:
            if ("hosts" not in env.sos_dict["CONFIG"] or
                    env.sos_dict["CONFIG"]["localhost"]
                    not in env.sos_dict["CONFIG"]["hosts"]):
                env.logger.warning(
                    f"Localhost {env.sos_dict['CONFIG']['localhost']} is not defined in CONFIG['hosts']"
                )
                env.sos_dict["CONFIG"]["hosts"][env.sos_dict["CONFIG"]
                                                ["localhost"]] = {
                                                    "paths": {},
                                                    "address": "localhost",
                                                }
            env.sos_dict.set("__host__", env.sos_dict["CONFIG"]["localhost"])
        else:
            if "hosts" in env.sos_dict["CONFIG"]:
                if "localhost" not in env.sos_dict["CONFIG"]["hosts"]:
                    env.logger.warning('locahost is not defined in "hosts".')
                    env.sos_dict["CONFIG"]["hosts"]["localhost"] = {
                        "paths": {},
                        "address": "localhost",
                    }
            elif "paths" not in env.sos_dict["CONFIG"]["hosts"]["localhost"]:
                env.sos_dict["CONFIG"]["hosts"]["localhost"]["paths"] = {}
            env.sos_dict.set("__host__", "localhost")
    else:
        if ("hosts" not in env.sos_dict["CONFIG"] or
                host not in env.sos_dict["CONFIG"]["hosts"]):
            raise RuntimeError(
                f"Remote host {host} is not defined in CONFIG['hosts']. Available ones are {env.sos_dict['CONFIG']['hosts'].keys()}"
            )
        env.sos_dict.set("__host__", host)


def statementMD5(stmts):

    def _get_tokens(statement):
        return [
            x[1]
            for x in generate_tokens(StringIO(statement).readline)
            if x[1] not in ("", "\n")
        ]

    tokens = []
    for stmt in stmts:
        if stmt:
            tokens.extend(_get_tokens(stmt))
    return textMD5(" ".join(tokens))


def create_task(global_def, global_vars, task_stmt, task_params):
    # env.sos_dict.set('_runtime', {})
    if task_params:
        args, kwargs = SoS_eval(
            f"__null_func__({task_params})",
            extra_dict={"__null_func__": __null_func__})
        if args:
            raise RuntimeError(
                f'Only keyword arguments are accepted for task statement: "{task_params}" provided'
            )
        for k, v in kwargs.items():
            # if k not in SOS_RUNTIME_OPTIONS:
            #     raise RuntimeError(f'Unrecognized runtime option {k}={v}')
            # standardize walltime to an integer
            if k == "walltime":
                v = format_HHMMSS(v)
            elif k == "mem":
                v = expand_size(v)
            env.sos_dict["_runtime"][k] = v
    #
    # we need to record the verbosity and sigmode of task during creation because
    # they might be changed while the task is in the queue waiting to be
    # submitted (this happens when tasks are submitted from Jupyter)
    env.sos_dict["_runtime"]["verbosity"] = env.verbosity
    env.sos_dict["_runtime"]["sig_mode"] = env.config.get("sig_mode", "default")
    env.sos_dict["_runtime"]["run_mode"] = env.config.get("run_mode", "run")

    # 1324
    env.sos_dict["_runtime"].update(env.config.get("queue_args", {}))

    # NOTE: we do not explicitly include 'step_input', 'step_output',
    # 'step_depends' and 'CONFIG'
    # because they will be included by env.sos_dict['__signature_vars__'] if they are actually
    # used in the task. (issue #752)
    task_vars = env.sos_dict.clone_selected_vars(
        env.sos_dict["__signature_vars__"]
        | {
            "_input",
            "_output",
            "_depends",
            "_index",
            "step_name",
            "_runtime",
            "__signature_vars__",
        })

    task_tags = [
        env.sos_dict["step_name"],
        env.sos_dict["workflow_id"],
        os.path.basename(os.getcwd()),
    ]
    if "tags" in env.sos_dict["_runtime"]:
        if isinstance(env.sos_dict["_runtime"]["tags"], str):
            tags = [env.sos_dict["_runtime"]["tags"]]
        elif isinstance(env.sos_dict["_runtime"]["tags"], Sequence):
            tags = [str(x) for x in list(env.sos_dict["_runtime"]["tags"])]
        else:
            tags = str(env.sos_dict["_runtime"]["tags"]).split(" ")
        #
        for tag in tags:
            if not tag.strip():
                continue
            if not SOS_TAG.match(tag):
                new_tag = re.sub(r"[^\w_.-]", "", tag)
                if new_tag:
                    env.logger.warning(
                        f'Invalid tag "{tag}" is added as "{new_tag}"')
                    task_tags.append(new_tag)
                else:
                    env.logger.warning(f'Invalid tag "{tag}" is ignored')
            else:
                task_tags.append(tag)

    # save task to a file
    taskdef = TaskParams(
        name="{} (index={})".format(env.sos_dict["step_name"],
                                    env.sos_dict["_index"]),
        global_def=(global_def, global_vars),
        task=task_stmt,  # task
        sos_dict=task_vars,
        tags=task_tags,
    )
    # if no output (thus no signature)
    # temporarily create task signature to obtain sig_id
    task_id = 't' + RuntimeInfo(
        statementMD5([task_stmt]),
        task_vars["_input"],
        task_vars["_output"],
        task_vars["_depends"],
        task_vars["__signature_vars__"],
        task_vars,
    ).sig_id

    # workflow ID should be included but not part of the signature, this is why it is included
    # after task_id is created.
    task_vars["workflow_id"] = env.sos_dict["workflow_id"]
    return task_id, taskdef, task_vars


def kill_all_subprocesses(pid=None, include_self=False):
    # kill all subprocesses that could have been spawn from the current process
    try:
        proc = psutil.Process(pid)
    except Exception:
        # if no such process
        return
    procs = proc.children(recursive=True) + ([proc] if include_self else [])
    if not procs:
        return
    for p in procs:
        p.terminate()
    alive = psutil.wait_procs(procs, timeout=3)[-1]
    if alive:
        for p in alive:
            p.kill()
    alive = psutil.wait_procs(procs, timeout=3)[-1]
    if alive:
        for p in alive:
            env.logger.warning(f"Failed to kill subprocess {p.pid}")


def reevaluate_output():
    # re-process the output statement to determine output files
    args, kwargs = SoS_eval(
        f'__null_func__({env.sos_dict["step_output"]._undetermined})',
        extra_dict={
            "__null_func__": __null_func__,
            "output_from": __output_from__,
            "named_output": __named_output__,
        },
    )
    if args is True:
        env.logger.error("Failed to resolve unspecified output")
        return
    # handle dynamic args
    args = [x.resolve() if isinstance(x, dynamic) else x for x in args]
    kwargs = {
        x: (y.resolve() if isinstance(y, dynamic) else y)
        for x, y in kwargs.items()
    }
    return sos_targets(*args, **kwargs, _verify_existence=True)


def validate_step_sig(sig):
    if env.config["sig_mode"] in ("default", "skip", "distributed"):
        # if users use sos_run, the "scope" of the step goes beyong names in this step
        # so we cannot save signatures for it.
        matched = sig.validate()
        if isinstance(matched, dict):
            env.logger.info(
                f'``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` due to saved signature'
            )
            return matched
        else:
            env.logger.debug(f"Signature mismatch: {matched}")
            return {}
    elif env.config["sig_mode"] == "assert":
        matched = sig.validate()
        if isinstance(matched, str):
            raise RuntimeError(f"Signature mismatch: {matched}")
        env.logger.info(
            f'Substep ``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` with matching signature'
        )
        return matched
    elif env.config["sig_mode"] == "build":
        # build signature require existence of files
        if sig.write():
            env.logger.info(
                f'Step ``{env.sos_dict["step_name"]}`` (index={env.sos_dict["_index"]}) is ``ignored`` with signature constructed'
            )
            return {
                "input": sig.content["input_obj"],
                "output": sig.content["output_obj"],
                "depends": sig.content["depends_obj"],
                "vars": sig.content["end_context"],
            }
    elif env.config["sig_mode"] == "force":
        return {}
    else:
        raise RuntimeError(
            f'Unrecognized signature mode {env.config["sig_mode"]}')


def strip_param_defs(stmt):
    # the parameters are translated to
    #
    # #begin_parameter name
    # name = sos_get_param("name", value
    # ) #end_parameter name
    #
    # we will need to remove these lines in cases when parameters are not handled
    res = []
    end_line = None
    for line in stmt.splitlines():
        if line.startswith("#begin_parameter "):
            end_line = f") #end_parameter {line[17:]}"
        if not end_line:
            res.append(line)
        elif line == end_line:
            end_line = None
    return "\n".join(res)


def verify_input(ignore_internal_targets=False):
    # now, if we are actually going to run the script, we
    # need to check the input files actually exists, not just the signatures
    for key in ("_input", "_depends"):
        for target in env.sos_dict[key]:
            if not target.target_exists("target") and not (
                    ignore_internal_targets and
                    isinstance(target, (sos_variable, sos_step))):
                raise RemovedTarget(target)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos.actions import SoS_Action, SoS_ExecuteScript


@SoS_Action(acceptable_args=["script", "args"])
def bash(script, args="", **kwargs):
    """Execute specified script using bash. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, ["/bin/bash", "bash"], ".sh",
                             args).run(**kwargs)


@SoS_Action(acceptable_args=["script", "args"])
def csh(script, args="", **kwargs):
    """Execute specified script using csh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, ["/bin/csh", "csh"], ".csh",
                             args).run(**kwargs)


@SoS_Action(acceptable_args=["script", "args"])
def tcsh(script, args="", **kwargs):
    """Execute specified script using tcsh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, ["/bin/tcsh", "tcsh"], ".sh",
                             args).run(**kwargs)


@SoS_Action(acceptable_args=["script", "args"])
def zsh(script, args="", **kwargs):
    """Execute specified script using zsh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, ["/bin/zsh", "zsh"], ".zsh",
                             args).run(**kwargs)


@SoS_Action(acceptable_args=["script", "args"])
def sh(script, args="", **kwargs):
    """Execute specified script using sh. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script."""
    return SoS_ExecuteScript(script, ["/bin/sh", "sh"], ".sh",
                             args).run(**kwargs)



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import argparse
import base64
import io

import pkg_resources

from sos.utils import dehtml, dot_to_gif, env, linecount_of_file


def get_previewers():
    # Note: data is zest.releaser specific: we want to pass
    # something to the plugin
    group = "sos_previewers"
    result = []
    for entrypoint in pkg_resources.iter_entry_points(group=group):
        # if ':' in entry point name, it should be a function
        try:
            name, priority = entrypoint.name.split(",", 1)
            priority = int(priority)
        except Exception as e:
            env.logger.warning(
                f"Ignore incorrect previewer entry point {entrypoint}: {e}")
            continue
        # If name points to a function in a module. Let us try to import the module
        if ":" in name:
            import importlib

            try:
                mod, func = name.split(":")
                imported = importlib.import_module(mod)
                result.append((getattr(imported, func), entrypoint, priority))
            except ImportError:
                env.logger.warning(f"Failed to load function {mod}:{func}")
        else:
            result.append((name, entrypoint, priority))
    #
    result.sort(key=lambda x: -x[2])
    # we put string infront of functions so that filename matching will be used before
    # content matching. For example, the xlsx file is actually a zip file so it could be
    # previewed as a zip file first.
    return ([x for x in result if isinstance(x[0], str) and x[0] != "*"] +
            [x for x in result if not isinstance(x[0], str)] +
            [x for x in result if x[0] == "*"])


def _preview_img_parser():
    parser = argparse.ArgumentParser(prog="%preview *.pdf")
    parser.add_argument(
        "--width",
        help="Width of the previewed image, can be in any HTML units such as px, em.",
    )
    parser.add_argument(
        "--height",
        help="Height of the previewed image, can be in any HTML units such as px, em.",
    )
    parser.error = lambda msg: env.logger.warning(msg)
    return parser


def preview_img(filename, kernel=None, style=None):
    with open(filename, "rb") as f:
        image = f.read()
    import imghdr

    image_type = imghdr.what(None, image)
    image_data = base64.b64encode(image).decode("ascii")

    args = None
    meta = {}
    if style is not None and "options" in style:
        parser = _preview_img_parser()
        try:
            args = parser.parse_args(style["options"])
            meta.update({
                "image/png":
                    dict(([["width", args.width]] if args.width else []) +
                         ([["height", args.height]] if args.height else []))
            })
        except SystemExit:
            return

    if image_type != "png":
        try:
            if image_type == "gif":
                return {"image/png": image_data}, meta
            else:
                from wand.image import Image

                img = Image(filename=filename)
                return {
                    "image/" + image_type:
                        image_data,
                    "image/png":
                        base64.b64encode(img._repr_png_()).decode("ascii"),
                }, meta
        except Exception:
            return {"image/" + image_type: image_data}, meta
    else:
        return {"image/" + image_type: image_data}, meta


def preview_svg(filename, kernel=None, style=None):
    with open(filename, "r") as f:
        image_data = f.read()
    return {"image/svg+xml": image_data}


def _preview_pdf_parser():
    parser = argparse.ArgumentParser(prog="%preview *.pdf")
    parser.add_argument(
        "--pages", nargs="+", type=int, help="Pages of the PDF to preview.")
    parser.add_argument(
        "--width",
        help="Width of the previewed image, can be in any HTML units such as px, em.",
    )
    parser.add_argument(
        "--height",
        help="Height of the previewed image, can be in any HTML units such as px, em.",
    )
    parser.add_argument(
        "--dpi",
        type=int,
        default=150,
        help="resolution of the converted png preview")
    parser.error = lambda msg: env.logger.warning(msg)
    return parser


def preview_pdf(filename, kernel=None, style=None):
    use_png = False
    warn = kernel.warn if kernel is not None else env.logger.warning
    if style is not None and "style" in style:
        if style["style"] != "png":
            if style["style"] is not None:
                warn(
                    f'Option --style of PDF preview only accept parameter png: {style["style"]} provided'
                )
        else:
            use_png = True
    args = None
    if style is not None and "options" in style:
        parser = _preview_pdf_parser()
        try:
            args = parser.parse_args(style["options"])
        except SystemExit:
            return
    meta = {}
    embed_options = ""
    if args and (args.width or args.height):
        meta.update({
            "image/png":
                dict(([["width", args.width]] if args.width else []) +
                     ([["height", args.height]] if args.height else []))
        })
        embed_options += (f'width="{args.width}" ' if args.width else " ") + (
            f'height="{args.height}" ' if args.height else " ")
    if use_png:
        try:
            # this import will fail even if wand is installed
            # if imagemagick is not installed properly.
            from wand.image import Image

            img = Image(filename=filename, resolution=args.dpi)

            if img.width == 0 or img.height == 0:
                raise ValueError("Image appears to have zero width or height")
            nPages = len(img.sequence)
            pages = list(range(nPages))

            if args and args.pages is not None:
                pages = [x - 1 for x in args.pages]
                for p in pages:
                    if p >= nPages:
                        warn(
                            f"Page {p} out of range of the pdf file ({nPages} pages)"
                        )
                        pages = list(range(nPages))
                        break
            # single page PDF
            if len(pages) == 1 and pages[0] == 0:
                return {
                    "image/png":
                        base64.b64encode(img._repr_png_()).decode("ascii")
                }, meta
            elif len(pages) == 1:
                # if only one page
                return {
                    "image/png":
                        base64.b64encode(
                            Image(img.sequence[pages[0]])._repr_png_()
                        ).decode("ascii")
                }, meta
            else:
                widths = [img.sequence[p].width for p in pages]
                heights = [img.sequence[p].height for p in pages]
                image = Image(width=max(widths), height=sum(heights))
                for i, p in enumerate(pages):
                    image.composite(
                        img.sequence[p], top=sum(heights[:i]), left=0)
                image.format = "png"
                with io.BytesIO() as out:
                    image.save(file=out)
                    img_data = out.getvalue()
                return {
                    "image/png": base64.b64encode(img_data).decode("ascii")
                }, meta
        except Exception as e:
            warn(e)
            return {
                "text/html":
                    f'<embed src="{filename}" {embed_options} type="application/pdf" />'
            }
    else:
        # by default use iframe, because PDF figure can have multiple pages (#693)
        return {
            "text/html":
                f'<embed src="{filename}" {embed_options} type="application/pdf" />'
        }


def preview_html(filename, kernel=None, style=None):
    with open(filename) as html:
        content = html.read()
    return {"text/html": content, "text/plain": dehtml(content)}


def _get_txt_parser():
    parser = argparse.ArgumentParser(prog="%preview *.txt")
    parser.add_argument(
        "-l",
        "--limit",
        type=int,
        default=5,
        help="""Maximum number
        of lines to preview.""",
    )
    parser.error = lambda msg: env.logger.warning(msg)
    return parser


def preview_txt(filename, kernel=None, style=None):
    if style is not None and "options" in style:
        parser = _get_txt_parser()
        try:
            args = parser.parse_args(style["options"])
        except SystemExit:
            return
        limit = args.limit
    else:
        limit = 5

    nlines = linecount_of_file(filename)
    content = "HINT: {} line{}{}\n".format(
        nlines,
        "s" if nlines > 1 else "",
        f" ({limit} displayed, see --limit)" if nlines > limit else "",
    )

    with open(filename, "r") as fin:
        if limit < 0:
            content += fin.read()
        else:
            for _ in range(limit):
                content += fin.readline()
    return content


def preview_csv(filename, kernel=None, style=None):
    import pandas

    from .visualize import Visualizer

    data = pandas.read_csv(filename)
    return Visualizer(kernel, style).preview(data)


def preview_xls(filename, kernel=None, style=None):
    import pandas

    from .visualize import Visualizer

    data = pandas.read_excel(filename)
    return Visualizer(kernel, style).preview(data)


def preview_zip(filename, kernel=None, style=None):
    import zipfile

    names = zipfile.ZipFile(filename).namelist()
    return (f"{len(names)} files\n" + "\n".join(names[:5]) +
            ("\n..." if len(names) > 5 else ""))


def preview_tar(filename, kernel=None, style=None):
    import tarfile

    with tarfile.open(filename, "r:*") as tar:
        # only extract files
        names = [x.name for x in tar.getmembers() if x.isfile()]
    return (f"{len(names)} files\n" + "\n".join(names[:5]) +
            ("\n..." if len(names) > 5 else ""))


def preview_gz(filename, kernel=None, style=None):
    import tarfile

    # .tar.gz
    if tarfile.is_tarfile(filename):
        return preview_tar(filename, kernel, style)
    import gzip

    content = b""
    with gzip.open(filename, "rb") as fin:
        for _ in range(5):
            content += fin.readline()
    try:
        return content.decode()[:2000]
    except Exception:
        return "binary data"


def preview_md(filename, kernel=None, style=None):
    import markdown

    with open(filename) as fin:
        text = fin.read()
    html = markdown.markdown(text)
    return {"text/html": html, "text/plain": text}


def preview_dot(filename, kernel=None, style=None):
    data = dot_to_gif(
        filename, warn=kernel.warn if kernel else env.logger.warning)
    # according to https://github.com/ipython/ipython/issues/10045
    # I have to use 'image/png' instead of 'image/gif' to get the gif displayed.
    return {"image/png": data}
